import importlib
import json
from copy import deepcopy

import jsonpath_rw as jsonpath

from .utils import ParamDefinitionException, ParamAdaptException


class ParamType(object):

    def __init__(self, name, value):
        assert isinstance(value, self.TYPES)
        self.name = name
        self.value = value

    @classmethod
    def is_valid(cls, value):
        return isinstance(value, cls.TYPES)


class Number(ParamType):
    TYPES = (int, float)
    DEFAULT = 0

    def __init__(self, name, value=None):
        value = 0 if value is None else value
        super().__init__(name, value)


class String(ParamType):
    TYPES = (str,)

    def __init__(self, name, value=None):
        value = '' if value is None else value
        super().__init__(name, value)


class Boolean(ParamType):
    TYPES = (bool,)

    def __init__(self, name, value=None):
        value = False if value is None else value
        super().__init__(name, value)


class Object(ParamType):
    TYPES = (dict,)

    def __init__(self, name, value=None):
        value = dict() if value is None else value
        super().__init__(name, value)


class Array(ParamType):
    TYPES = (list,)

    def __init__(self, name, value=None):
        value = list() if value is None else value
        super().__init__(name, value)


def get_param_class(cls_name):
    """
    """

    cls = getattr(importlib.import_module(__name__), cls_name)
    assert issubclass(cls, ParamType)
    return cls


class ParamDefinitionItem(object):
    def __init__(self):
        self.type = None
        self.required = None
        self.default = None

    @classmethod
    def from_json(cls, definition_item: dict):
        """
        :param definition_item:
        :return:
        """
        _type = get_param_class(definition_item.get('type'))
        required = definition_item.get('required', True)
        default = definition_item.get('default', None)
        assert issubclass(_type, ParamType)
        if required and default is not None:
            assert _type.is_valid(default)
        r = cls()
        r.type = _type
        r.required = required
        r.default = default

        return r

    def to_json(self):
        return {'type': self.type.__name__, 'required': self.required, 'default': self.default}


class ParamDefinition(object):
    def __init__(self):
        self.dict = {}

    def __len__(self):
        return len(self.dict)

    def add(self, name, item: ParamDefinitionItem):
        assert isinstance(item, ParamDefinitionItem)
        self.dict[name] = item

    def get(self, name):
        return self.dict[name]

    def has(self, name):
        return name in self.dict

    def keys(self):
        return self.dict.keys()

    def dump(self):
        return json.dumps(self.to_json(), ensure_ascii=False)

    @classmethod
    def from_json(cls, definition: dict):
        """
        :param definition: dict, {"paramA": "Number"}
        :return:
        """
        r = cls()
        for name, item in definition.items():
            r.add(name, ParamDefinitionItem.from_json(item))
        return r

    def to_json(self):
        """
        :return:
        """
        d = {k: v.to_json() for k, v in self.dict.items()}
        return d

    def validate(self, params):
        """
        :param params: dict
        :return:
        """
        errors = {}
        for name, item in self.dict.items():
            if not params or params.get(name) is None:
                if item.required:
                    errors[name] = 'missing required field'
            else:
                if not item.type.is_valid(params[name]):
                    errors[name] = 'invalid value, expect %s' % item.type.__name__

        return len(errors) == 0, errors


class ParamAdapter(object):
    def __init__(self):
        self.dict = {}
        self.parser_dict = {}

    def __len__(self):
        return len(self.dict)

    def parse(self, raw):
        return self.adapt(raw)

    def adapt(self, raw):
        """
        :param raw: dict, raw params
        :return:
        """

        d = {}

        raw = raw or {}
        def __recursive(parser, o, key):
            if isinstance(parser, dict):
                o[key] = {}
                for k, v in parser.items():
                    __recursive(v, o[key], k)
            else:
                try:
                    rs = parser.find(raw)
                    if rs:
                        assert len(rs) == 1, 'duplicated hits'
                        o[key] = rs[0].value
                except Exception as e:
                    raise ParamAdaptException(key).with_traceback(e.__traceback__)

        for name, item in self.parser_dict.items():
            __recursive(item, d, name)

        return d


    def dump(self):
        return json.dumps(self.dict, ensure_ascii=False)

    @classmethod
    def from_json(cls, adapter: dict):
        """
        :param adapter: dict, {"paramA": "$.a"}
        :return:
        """
        def __recursive(o):
            for k, v in o.items():
                if isinstance(v, dict):
                    __recursive(v)
                else:
                    o[k] = jsonpath.parse(v)
            return o

        r = cls()
        r.dict = adapter
        r.parser_dict = __recursive(deepcopy(adapter))

        return r

    def to_json(self):
        """
        :return:
        """
        d = {k: v for k, v in self.dict.items()}
        return d


if __name__ == '__main__':
    pd = ParamDefinition.from_json({'a': 'Array', 'b': 'Object', 'c': 'Number'})
    adapter = ParamAdapter.from_json({'a': '$.aa', 'b': '$.bb', 'c': '$.cc'})
    r = adapter.parse({'aa': [1, 2], 'bb': {'x': 1}, 'cc': 12})
    print(r, pd.validate(r))



# Generated by Django 3.1 on 2022-07-22 04:55

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Action',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False)),
                ('name', models.CharField(db_index=True, max_length=64, unique=True)),
                ('title', models.CharField(max_length=64)),
                ('type', models.CharField(choices=[('Default', 'default'), ('Carrier', 'carrier'), ('External', 'external')], max_length=32)),
                ('func', models.CharField(max_length=128, null=True, unique=True)),
                ('input_def', models.JSONField(default=dict)),
                ('output_def', models.JSONField(default=dict)),
                ('create_time', models.DateTimeField(auto_now_add=True)),
                ('update_time', models.DateTimeField(auto_now=True)),
            ],
            options={
                'verbose_name': '流程Action',
                'verbose_name_plural': '流程Action',
                'db_table': 'seaflow_action',
                'managed': True,
            },
        ),
        migrations.CreateModel(
            name='Dag',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False)),
                ('name', models.CharField(db_index=True, max_length=64)),
                ('title', models.CharField(max_length=64)),
                ('version', models.PositiveIntegerField(db_index=True, null=True)),
                ('latest', models.BooleanField(null=True)),
                ('input_adapter', models.JSONField(default=dict)),
                ('output_adapter', models.JSONField(default=dict)),
                ('fissionable', models.BooleanField(default=False, verbose_name='可分裂')),
                ('fission_config', models.JSONField(default=dict)),
                ('iterable', models.BooleanField(default=False, verbose_name='可迭代')),
                ('iter_config', models.JSONField(default=dict)),
                ('loopable', models.BooleanField(default=False, verbose_name='可循环')),
                ('loop_config', models.JSONField(default=dict)),
                ('max_retries', models.IntegerField(default=0)),
                ('retry_countdown', models.IntegerField(default=0)),
                ('timeout', models.IntegerField(null=True)),
                ('create_time', models.DateTimeField(auto_now_add=True)),
                ('update_time', models.DateTimeField(auto_now=True)),
                ('parent', models.ForeignKey(db_constraint=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='seaflow.dag')),
                ('previous_dags', models.ManyToManyField(db_constraint=False, db_table='seaflow_dag_to_dag', related_name='next_dags', to='seaflow.Dag')),
            ],
            options={
                'verbose_name': '流程图',
                'verbose_name_plural': '流程图',
                'db_table': 'seaflow_dag',
                'managed': True,
            },
        ),
        migrations.CreateModel(
            name='Log',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False)),
                ('ref_type', models.CharField(choices=[('TASK', 'Task'), ('STEP', 'Step')], max_length=16)),
                ('ref_id', models.IntegerField(db_index=True)),
                ('ts', models.FloatField()),
                ('content', models.TextField()),
            ],
            options={
                'verbose_name': '日志',
                'verbose_name_plural': '日志',
                'db_table': 'seaflow_log',
                'managed': True,
            },
        ),
        migrations.CreateModel(
            name='Node',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=64)),
                ('title', models.CharField(max_length=64)),
                ('input_adapter', models.JSONField(default=dict)),
                ('output_adapter', models.JSONField(default=dict)),
                ('action_type', models.CharField(choices=[('Default', 'default'), ('Carrier', 'carrier'), ('External', 'external')], max_length=32)),
                ('fissionable', models.BooleanField(default=False, verbose_name='可分裂')),
                ('fission_config', models.JSONField(default=dict)),
                ('iterable', models.BooleanField(default=False, verbose_name='可迭代')),
                ('iter_config', models.JSONField(default=dict)),
                ('loopable', models.BooleanField(default=False, verbose_name='可循环')),
                ('loop_config', models.JSONField(default=dict)),
                ('max_retries', models.IntegerField(default=0)),
                ('retry_countdown', models.IntegerField(default=0)),
                ('timeout', models.IntegerField(null=True)),
                ('create_time', models.DateTimeField(auto_now_add=True)),
                ('update_time', models.DateTimeField(auto_now=True)),
                ('action', models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.PROTECT, to='seaflow.action')),
                ('dag', models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.CASCADE, related_name='nodes', to='seaflow.dag')),
                ('previous_dags', models.ManyToManyField(db_constraint=False, db_table='seaflow_dag_to_node', related_name='next_nodes', to='seaflow.Dag')),
                ('previous_nodes', models.ManyToManyField(db_constraint=False, db_table='seaflow_node_to_node', related_name='next_nodes', to='seaflow.Node')),
                ('root_dag', models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.CASCADE, related_name='all_nodes', to='seaflow.dag')),
            ],
            options={
                'verbose_name': '流程节点',
                'verbose_name_plural': '流程节点',
                'db_table': 'seaflow_node',
                'managed': True,
            },
        ),
        migrations.CreateModel(
            name='Step',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False)),
                ('identifier', models.CharField(db_index=True, max_length=64, null=True)),
                ('name', models.CharField(max_length=64)),
                ('title', models.CharField(max_length=64)),
                ('state', models.CharField(choices=[('PENDING', '等待中'), ('PUBLISH', '已发布'), ('PROCESSING', '正在执行'), ('SLEEP', '休眠'), ('RETRY', '等待重试'), ('TIMEOUT', '超时'), ('SUCCESS', '成功'), ('ERROR', '错误'), ('REVOKE', '撤销'), ('TERMINATE', '终止')], max_length=64)),
                ('config', models.JSONField(default=dict, verbose_name='配置')),
                ('retries', models.IntegerField(default=0, verbose_name='重试次数')),
                ('fission_count', models.IntegerField(default=1, verbose_name='分裂数')),
                ('fission_index', models.IntegerField(db_index=True, default=0, verbose_name='分裂序号')),
                ('iter_index', models.IntegerField(db_index=True, default=0, verbose_name='迭代序号')),
                ('iter_end', models.BooleanField(null=True, verbose_name='迭代末端')),
                ('loop_index', models.IntegerField(default=0, verbose_name='循环序号')),
                ('input', models.JSONField(verbose_name='输入')),
                ('output', models.JSONField(default=dict, verbose_name='输出')),
                ('extra', models.JSONField(default=dict)),
                ('error', models.TextField(default='', verbose_name='错误信息')),
                ('logs', models.JSONField(default=list, verbose_name='执行日志')),
                ('logfile', models.CharField(default='', max_length=256, verbose_name='s3日志文件路径')),
                ('log_refetch_time', models.DateTimeField(db_index=True, null=True)),
                ('start_time', models.DateTimeField(null=True)),
                ('end_time', models.DateTimeField(null=True)),
                ('duration', models.FloatField(null=True)),
                ('create_time', models.DateTimeField(auto_now_add=True)),
                ('update_time', models.DateTimeField(auto_now=True)),
                ('node', models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.PROTECT, to='seaflow.node')),
                ('previous_steps', models.ManyToManyField(db_constraint=False, db_table='seaflow_step_to_step', related_name='next_steps', to='seaflow.Step')),
            ],
            options={
                'verbose_name': '任务步骤',
                'verbose_name_plural': '任务步骤',
                'db_table': 'seaflow_step',
                'managed': True,
            },
        ),
        migrations.CreateModel(
            name='Task',
            fields=[
                ('id', models.AutoField(primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=64)),
                ('title', models.CharField(max_length=64)),
                ('state', models.CharField(choices=[('PENDING', '等待中'), ('PROCESSING', '正在执行'), ('SLEEP', '休眠'), ('RETRY', '等待重试'), ('TIMEOUT', '超时'), ('SUCCESS', '成功'), ('ERROR', '错误'), ('REVOKE', '撤销'), ('TERMINATE', '终止')], max_length=64)),
                ('config', models.JSONField(default=dict, verbose_name='配置')),
                ('retries', models.IntegerField(default=0, verbose_name='重试次数')),
                ('fission_count', models.IntegerField(default=1, verbose_name='分裂数')),
                ('fission_index', models.IntegerField(db_index=True, default=0, verbose_name='分裂序号')),
                ('iter_index', models.IntegerField(db_index=True, default=0, verbose_name='迭代序号')),
                ('iter_end', models.BooleanField(null=True, verbose_name='迭代末端')),
                ('loop_index', models.IntegerField(default=0, verbose_name='循环序号')),
                ('input', models.JSONField(verbose_name='输入')),
                ('context', models.JSONField(default=dict, verbose_name='上下文')),
                ('output', models.JSONField(default=dict, verbose_name='输出')),
                ('extra', models.JSONField(default=dict)),
                ('error', models.TextField(default='', verbose_name='错误信息')),
                ('logs', models.JSONField(default=list, verbose_name='执行日志')),
                ('logfile', models.CharField(default='', max_length=256, verbose_name='s3日志文件路径')),
                ('log_refetch_time', models.DateTimeField(db_index=True, null=True)),
                ('start_time', models.DateTimeField(null=True)),
                ('end_time', models.DateTimeField(null=True)),
                ('duration', models.FloatField(null=True)),
                ('create_time', models.DateTimeField(auto_now_add=True)),
                ('update_time', models.DateTimeField(auto_now=True)),
                ('dag', models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.PROTECT, to='seaflow.dag')),
                ('parent', models.ForeignKey(db_constraint=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='children', to='seaflow.task')),
                ('previous_steps', models.ManyToManyField(db_constraint=False, db_table='seaflow_step_to_task', related_name='next_tasks', to='seaflow.Step')),
                ('previous_tasks', models.ManyToManyField(db_constraint=False, db_table='seaflow_task_to_task', related_name='next_tasks', to='seaflow.Task')),
                ('root', models.ForeignKey(db_constraint=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='descendants', to='seaflow.task')),
            ],
            options={
                'verbose_name': '任务',
                'verbose_name_plural': '任务',
                'db_table': 'seaflow_task',
                'managed': True,
                'unique_together': {('parent', 'dag', 'fission_index', 'iter_index')},
            },
        ),
        migrations.AddField(
            model_name='step',
            name='previous_tasks',
            field=models.ManyToManyField(db_constraint=False, db_table='seaflow_task_to_step', related_name='next_steps', to='seaflow.Task'),
        ),
        migrations.AddField(
            model_name='step',
            name='root',
            field=models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.CASCADE, related_name='all_steps', to='seaflow.task'),
        ),
        migrations.AddField(
            model_name='step',
            name='task',
            field=models.ForeignKey(db_constraint=False, on_delete=django.db.models.deletion.CASCADE, related_name='steps', to='seaflow.task'),
        ),
        migrations.AddField(
            model_name='dag',
            name='previous_nodes',
            field=models.ManyToManyField(db_constraint=False, db_table='seaflow_node_to_dag', related_name='next_dags', to='seaflow.Node'),
        ),
        migrations.AddField(
            model_name='dag',
            name='root',
            field=models.ForeignKey(db_constraint=False, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='descendants', to='seaflow.dag'),
        ),
        migrations.AlterUniqueTogether(
            name='step',
            unique_together={('task', 'node', 'fission_index', 'iter_index')},
        ),
    ]



import json

from seaflow.consts import StepStates
from seaflow.base import Seaflow

print(1)
@Seaflow.action()
def test(self, **params):
    print(json.dumps({'root': self.root.name if self.root else None,
                      'task': self.task.name,
                      'step': self.step.name}))
    if self.step.node.name == 'd1' and self.step.retries < self.step.config.get('max_retries', 0):
        1/0
    return {
        'state': StepStates.SUCCESS,
        'data': {
            'a': [1, 2],
            'o': {'o': 'o'},
            'n': 1,
            's': 's',
            'b': True
        }
    }



from enum import unique

from .utils import NameComparableEnum


@unique
class TaskStates(NameComparableEnum):
    PENDING = '等待中'
    PROCESSING = '正在执行'
    SLEEP = '休眠'
    RETRY = '等待重试'
    TIMEOUT = '超时'
    SUCCESS = '成功'
    ERROR = '错误'
    REVOKE = '撤销'
    TERMINATE = '终止'


@unique
class StepStates(NameComparableEnum):
    PENDING = '等待中'
    PUBLISH = '已发布' # 外部节点
    PROCESSING = '正在执行'
    SLEEP = '休眠'
    RETRY = '等待重试'
    TIMEOUT = '超时'
    SUCCESS = '成功'
    ERROR = '错误'
    REVOKE = '撤销'
    TERMINATE = '终止'


@unique
class ActionTypes(NameComparableEnum):
    Default = 'default'
    Carrier = 'carrier'
    External = 'external'



import functools
import hashlib
import json
import sys
from copy import deepcopy

from celery.result import AsyncResult
from django.db import transaction, models
from django.utils import timezone
from json_logic import jsonLogic
from timeout_decorator import timeout

from . import celery_app, tasks
from .consts import ActionTypes, TaskStates, StepStates
from .models import Action, Dag, Node, Task, Step
from .params import ParamAdapter, ParamDefinition
from .seagull import Seagull
from .tasks import do_callback
from .utils import *


class Seaflow(object):

    @classmethod
    def create_task(cls, *args, **kwargs):
        return SeaflowTask.create(*args, **kwargs)

    @classmethod
    def get_task(cls, task_id):
        return SeaflowTask.get(task_id)

    @staticmethod
    @transaction.atomic()
    def load_actions(dsl):
        """
        load actions from dsl
        :param dsl: json dsl
        :return:
        """

        actions = []
        for item in dsl:
            actions.append(Action(
                name=item['name'],
                title=item.get('title', item['name']),
                func=item.get('func'),
                type=item.get('type', ActionTypes.Default),
                input_def=item.get('input_def', {}),
                output_def=item.get('output_def', {}),
            ))
        with transaction.atomic():
            for a in actions:
                a.save()
            # assert len(actions) == len(dsl), 'not all actions are created'
        return actions

    @staticmethod
    @transaction.atomic()
    def load_dag(dsl):
        """
        load dag from dsl
        :param dsl: json dsl
        :return:
        """
        dsl = deepcopy(dsl)
        root_identifier = dsl['identifier']
        d = {
            root_identifier: dsl,
        }

        # 先处理refs
        ref_components = []
        for item in dsl['components']:
            if item.get('ref'):
                assert item['kind'] == 'Dag'
                _ = item['ref'].rsplit('.', 1)
                ref_name, ref_version = _[0], None if len(_) == 1 else _[1]
                ref = Seaflow.dag_dsl(dag_name=ref_name, dag_version=ref_version, root_identifier=item['identifier'])
                ref_components += ref['components']

        dsl['components'] += ref_components

        action_dict = {}
        node_dict = {}
        dag_children_dict = {}
        for item in dsl['components']:
            assert item['identifier'] not in d, 'duplicated identifier %s' % item['identifier']
            d[item['identifier']] = item
            if item['kind'] == 'Dag':
                parent = item.get('parent', root_identifier)
                if parent not in dag_children_dict:
                    dag_children_dict[parent] = [item['identifier']]
                else:
                    dag_children_dict[parent].append(item['identifier'])
            elif item['kind'] == 'Node':
                action_dict[item['action']] = None
                node_dict[item['identifier']] = item

        action_length = len(action_dict)
        action_dict = {a.name: a for a in Action.objects.filter(name__in=action_dict.keys())}
        assert action_length == len(action_dict), 'action not exist'

        def __recursive_create_dags(_identifiers):
            for x in _identifiers:
                dag_item = d[x]
                if x == root_identifier:
                    # root dag
                    version = dag_item.get('version', 1)
                    exists_dags = Dag.objects.filter(name=dag_item['name'], parent=None).order_by('-version')
                    latest_dag = None
                    if exists_dags: latest_dag = exists_dags[0]
                    assert latest_dag.version < version, 'out of version'
                    Dag.objects.filter(name=dag_item['name'], parent=None).update(latest=False)
                    d[x] = Dag.objects.create(
                        name=dag_item['name'],
                        title=dag_item.get('title', dag_item['name']),
                        version=version,
                        latest=True,
                        max_retries=dag_item.get('max_retries', 0),
                        timeout=dag_item.get('timeout'),
                        input_adapter=dag_item.get('input_adapter', {}),
                        output_adapter=dag_item.get('output_adapter', {})
                    )
                else:
                    d[x] = Dag.objects.create(
                        name=dag_item['name'],
                        title=dag_item.get('title', dag_item['name']),
                        root=d[root_identifier],
                        parent=d[dag_item.get('parent', root_identifier)],
                        fissionable=bool(dag_item.get('fission')),
                        fission_config=dag_item.get('fission', {}),
                        iterable=bool(dag_item.get('iter')),
                        iter_config=dag_item.get('iter', {}),
                        loopable=bool(dag_item.get('loop')),
                        loop_config=dag_item.get('loop', {}),
                        max_retries=dag_item.get('max_retries', 0),
                        timeout=dag_item.get('timeout'),
                        input_adapter=dag_item.get('input_adapter', {}),
                        output_adapter=dag_item.get('output_adapter', {})
                    )
                children = dag_children_dict.pop(x, [])
                if children:
                    __recursive_create_dags(children)

        with transaction.atomic():
            __recursive_create_dags([root_identifier])
            for k, node_item in node_dict.items():
                a = action_dict[node_item['action']]
                d[k] = Node.objects.create(
                    name=node_item['name'],
                    title=node_item.get('title', node_item['name']),
                    dag=d[node_item.get('dag', root_identifier)],
                    root_dag=d[root_identifier],
                    fissionable=bool(node_item.get('fission')),
                    fission_config=node_item.get('fission', {}),
                    iterable=bool(node_item.get('iter')),
                    iter_config=node_item.get('iter', {}),
                    loopable=bool(node_item.get('loop')),
                    loop_config=node_item.get('loop', {}),
                    action=a,
                    action_type=a.type,
                    max_retries=node_item.get('max_retries', 0),
                    timeout=node_item.get('timeout'),
                    input_adapter=node_item.get('input_adapter', {}),
                    output_adapter=node_item.get('output_adapter', {})
                )

            # relations
            for item in dsl['components']:
                if item.get('previous_dags'):
                    d[item['identifier']].previous_dags.set([d[x] for x in item['previous_dags']])
                if item.get('previous_nodes'):
                    d[item['identifier']].previous_nodes.set([d[x] for x in item['previous_nodes']])

        return d[root_identifier]

    @staticmethod
    def dump_dag_dsl(*args, **kwargs):
        dsl = Seaflow.dag_dsl(*args, **kwargs)
        return json.dumps(dsl, ensure_ascii=False, indent=2)

    @staticmethod
    def dag_dsl(dag_id=None, dag_name=None, dag_version=None, root_identifier=None):
        def _gen_identifier(s, prefix=root_identifier, use_hash=True):
            identifier = s
            if prefix:
                identifier = str(prefix) + str(s)
            if use_hash:
                return hashlib.md5(identifier.encode()).hexdigest()
            else:
                return identifier

        assert dag_id or dag_name
        try:
            if dag_id:
                dag = Dag.objects.prefetch_related('descendants', 'all_nodes', 'all_nodes__action').get(pk=dag_id)
            else:
                if dag_version:
                    dag = Dag.objects.prefetch_related('descendants', 'all_nodes', 'all_nodes__action').get(name=dag_name,
                                                                                                        version=dag_version)
                else:
                    dag = Dag.objects.prefetch_related('descendants', 'all_nodes', 'all_nodes__action').get(name=dag_name, latest=True)
        except models.ObjectDoesNotExist as e:
            sys.stderr.writelines(['dag not exists, id: %s, name: %s, version: %s' % (dag_id, dag_name, dag_version)])
            raise

        dag_identifiers = {
            dag.id: root_identifier or _gen_identifier(str(dag.id), prefix='root-%s-idx-' % dag.name, use_hash=False)
        }
        node_identifiers = {}
        for d in dag.descendants.all():
            dag_identifiers[d.id] = _gen_identifier(str(d.id), prefix='dag-%s-idx-' % d.name, use_hash=False)
        for n in dag.all_nodes.all():
            node_identifiers[n.id] = _gen_identifier(str(n.id), prefix='node-%s-idx-' % n.name, use_hash=False)
        dsl = {
            'identifier': dag_identifiers[dag.id],
            'name': dag.name,
            'title': dag.title,
            'version': dag.version,
            'latest': dag.latest,
            'input_adapter': dag.input_adapter,
            'output_adapter': dag.output_adapter,
            'components': []
        }

        for d in dag.descendants.all():
            d_data = {
                'identifier': dag_identifiers[d.id],
                'kind': 'Dag',
                'name': d.name,
                'title': d.title,
                'root': dag_identifiers[dag.id],
                'parent': dag_identifiers[d.parent.id],
                'input_adapter': d.input_adapter,
                'output_adapter': d.output_adapter,
                'previous_dags': [dag_identifiers[x.id] for x in d.previous_dags.all()],
                'previous_nodes': [node_identifiers[x.id] for x in d.previous_nodes.all()],
            }
            if d.fissionable: d_data['fission'] = d.fission_config
            if d.iterable: d_data['iter'] = d.iter_config
            if d.loopable: d_data['loop'] = d.loop_config
            dsl['components'].append(d_data)

        for n in dag.all_nodes.all():
            n_data = {
                'identifier': node_identifiers[n.id],
                'kind': 'Node',
                'name': n.name,
                'title': n.title,
                'dag': dag_identifiers[n.dag.id],
                'root_dag': dag_identifiers[dag.id],
                'fission': n.fission_config,
                'action': n.action.name,
                'action_type': n.action_type,
                'max_retries': n.max_retries,
                'input_adapter': n.input_adapter,
                'output_adapter': n.output_adapter,
                'previous_dags': [dag_identifiers[x.id] for x in n.previous_dags.all()],
                'previous_nodes': [node_identifiers[x.id] for x in n.previous_nodes.all()],
            }
            if n.fissionable: n_data['fission'] = n.fission_config
            if n.iterable: n_data['iter'] = n.iter_config
            if n.loopable: n_data['loop'] = n.loop_config
            dsl['components'].append(n_data)

        return dsl

    @classmethod
    def action(cls, *args, **opts):
        """
        :param args:
        :param opts:
        :return:
        """

        def _dec(func):
            @functools.wraps(func)
            def __dec(celery_action, step_id):
                celery_action.func = func
                SeaflowStep.get(step_id)._execute(celery_action)

            return celery_app.task(bind=True)(__dec)

        return _dec

    @classmethod
    def sleep_task(cls, task_id):
        """
        使任务进入睡眠状态
        :param task_id:
        :return:
        """

        SeaflowTask.get(task_id).sleep()

    @classmethod
    def awake_task(cls, task_id):
        """
        唤醒任务
        :param task_id:
        :return:
        """

        SeaflowTask.get(task_id).awake()

    @classmethod
    def revoke_task(cls, task_id):
        """
        撤销任务
        :param task_id:
        :return:
        """

        SeaflowTask.get(task_id).revoke()

    @classmethod
    def terminate_task(cls, task_id):
        """
        终止任务
        :param task_id:
        :return:
        """

        SeaflowTask.get(task_id).terminate()

    @classmethod
    def dispatch_external_step(cls, step_id, identity={}):
        """
        派发外部step
        :param step_id:
        :param identity: dict，订阅者的身份信息
                    name:
                    key:
        :return: identifier
        """
        return SeaflowStep.get(step_id)._dispatch(identity)

    @classmethod
    def finish_external_step(cls, step_id, outputs={}):
        """
        完成外部step
        :param step_id:
        :param outputs:
        :return:
        """
        step = SeaflowStep.get(step_id)
        if step.model.node.action_type != ActionTypes.External:
            raise SeaflowException('step is not external')
        try:
            step.seagull.info('action【%s】 output: %s' % (step.model.node.action.name,
                                                         json.dumps(outputs, cls=ComplexJSONEncoder)))
            step._finish(step._adapt_outputs(outputs))
        finally:
            step.seagull.flush(True)

    @classmethod
    def fail_external_step(cls, step_id, error=None, outputs={}):
        """
        失败外部step
        :param step_id:
        :param error:
        :param outputs:
        :return:
        """
        step = SeaflowStep.get(step_id)
        if step.model.node.action_type != ActionTypes.External:
            raise SeaflowException('step is not external')
        try:
            step.seagull.error('receive error: %s' % error)
            step.seagull.info('action【%s】 output: %s' % (step.model.node.action.name,
                                                         json.dumps(outputs, cls=ComplexJSONEncoder)))
            try:
                outputs = step._adapt_outputs(outputs)
            except ParamDefinitionException:
                outputs = {}
            step._break_off(e=ExternalActionFailed(error) if error else None, outputs=outputs)
        finally:
            step.seagull.flush(True)


class SeaflowTask(object):

    def __init__(self):
        # model
        self.model = None
        self.id = None
        self.name = None
        self.context = None
        self.seagull = None
        self.config = None

        #
        self.parent = None  # SeaflowTask
        self.root = None  # SeaflowTask

    @classmethod
    def create(cls, dag_id=None, dag_name=None, dag_version=None, name=None,
               inputs=None, context=None, config=None, tasks_config=None, steps_config=None):
        """
        :param dag_id:
        :param dag_name:
        :param dag_version:
        :param name:
        :param inputs:
        :param context:
        :param config: TaskConfig
        :param tasks_config: {<dag_name>: TaskConfig}
        :param steps_config: {<node_name>: StepConfig}
        :return: SeaflowTask
        """
        assert dag_id or dag_name
        if dag_id:
            dag = Dag.objects.get(pk=dag_id)
        else:
            if dag_version:
                dag = Dag.objects.get(name=dag_name, version=dag_version)
            else:
                dag = Dag.objects.get(name=dag_name, latest=True)
        assert not dag.root_id, 'assert to be root dag'
        inputs = ParamAdapter.from_json(dag.input_adapter).adapt(inputs or {})
        extra = {'tasks_config': tasks_config or {}, 'steps_config': steps_config or {}}
        t = Task.objects.create(
            name=name or dag.name,
            title=dag.title,
            dag=dag,
            state=TaskStates.PENDING,
            root=None,
            parent=None,
            input=inputs or {},
            context=context or {},
            config=config or {},
            extra=extra,
        )
        r = cls.get(task=t)
        r.seagull.info('task 【%s】 created: %s' % (t.name, t.id))
        r.seagull.flush(True)

        return r

    @classmethod
    def get(cls, task_id=None, task=None):
        """
        :param task_id:
        :param task:
        :return: SeaflowTask
        """
        m = task or Task.objects.select_related('dag', 'parent', 'root').defer('logs').get(pk=task_id)
        r = cls()
        r.model = m
        r.load()
        return r

    def load(self):
        self.id = self.model.id
        self.name = self.model.name
        self.context = (self.model.root or self.model).context
        self.seagull = Seagull.instance(self.model)
        self.config = self.model.config

        if self.model.parent:
            self.parent = self.__class__.get(task=self.model.parent)
        if self.model.root:
            self.root = self.__class__.get(task=self.model.root)

    def reload(self):
        self.model.refresh_from_db()
        self.load()

    def apply(self, sync=False, countdown=None):
        if sync:
            self._apply()
        else:
            tasks.apply_root_task.apply_async((self.id,), countdown=countdown)

    def _apply(self):
        self.model.update(start_time=timezone.now(), state=StepStates.PROCESSING)
        self._do_callback('TASK_STATE_%s' % TaskStates.PROCESSING)
        self.seagull.info('task 【%s】 started: %s' % (self.name, self.id))

        # 第一批free node
        nodes = self.model.dag.nodes.filter(previous_dags=None, previous_nodes=None)
        if nodes:
            self.seagull.info('apply nodes...')
            [self._apply_node(n) for n in nodes]

        # 第一批dag
        dags = self.model.dag.children.filter(previous_dags=None, previous_nodes=None)
        if dags:
            self.seagull.info('apply dags...')
            [self._apply_dag(d) for d in dags]

        self.seagull.flush(True)

    def revoke(self):
        self.reload()
        if self.model.state in [TaskStates.PENDING, TaskStates.PROCESSING, TaskStates.RETRY]:
            tnow = timezone.now()
            duration = 0
            if self.model.start_time:
                duration = tnow - self.model.start_time
                duration = float('%s.%s' % (duration.seconds, duration.microseconds))
            self.model.update(
                state=TaskStates.REVOKE,
                end_time=tnow,
                duration=duration
            )
            self._do_callback('TASK_STATE_%s' % TaskStates.REVOKE)
        for t in self.model.descendants.filter(state__in=[TaskStates.PENDING, TaskStates.PROCESSING, TaskStates.RETRY]):
            tnow = timezone.now()
            duration = 0
            if t.start_time:
                duration = tnow - t.start_time
                duration = float('%s.%s' % (duration.seconds, duration.microseconds))
            t.update(
                state=TaskStates.REVOKE,
                end_time=tnow,
                duration=duration
            )

    def terminate(self):
        self.reload()
        if self.model.state in [TaskStates.PENDING, TaskStates.PROCESSING, TaskStates.RETRY, TaskStates.REVOKE]:
            tnow = timezone.now()
            duration = 0
            if self.model.start_time:
                duration = tnow - self.model.start_time
                duration = float('%s.%s' % (duration.seconds, duration.microseconds))
            self.model.update(
                state=TaskStates.TERMINATE,
                end_time=timezone.now(),
                duration=duration
            )
            self._do_callback('TASK_STATE_%s' % TaskStates.TERMINATE)
        for t in self.model.descendants.filter(state__in=[TaskStates.PENDING, TaskStates.PROCESSING,
                                                          TaskStates.RETRY, TaskStates.REVOKE]):
            tnow = timezone.now()
            duration = 0
            if t.start_time:
                duration = tnow - t.start_time
                duration = float('%s.%s' % (duration.seconds, duration.microseconds))
            Seagull.instance(t).flush(True, merge=True)
            t.update(
                state=TaskStates.TERMINATE,
                end_time=tnow,
                duration=duration
            )
        for step in self.model.all_steps.filter(state__in=[StepStates.PENDING,
                                                           StepStates.PUBLISH,
                                                           StepStates.PROCESSING,
                                                           StepStates.SLEEP,
                                                           StepStates.RETRY]):
            tnow = timezone.now()
            duration = 0
            if step.start_time:
                duration = tnow - step.start_time
                duration = float('%s.%s' % (duration.seconds, duration.microseconds))
            Seagull.instance(step).flush(True, merge=True)
            step.update(
                state=StepStates.TERMINATE,
                end_time=timezone.now(),
                duration=duration
            )
            if AsyncResult(step.identifier).state in ['PENDING', 'RECEIVED', 'STARTED', 'RETRY']:
                celery_app.control.revoke(step.identifier, terminate=True)

    def sleep(self):
        self.reload()
        if self.model.state in [TaskStates.PENDING, TaskStates.PROCESSING, TaskStates.RETRY]:
            self.model.update(state=TaskStates.SLEEP)
            self._do_callback('TASK_STATE_%s' % TaskStates.SLEEP)
        for t in self.model.descendants.filter(state__in=[TaskStates.PENDING, TaskStates.PROCESSING, TaskStates.RETRY]):
            t.update(state=TaskStates.SLEEP)
            SeaflowTask.get(task=t)._do_callback('TASK_STATE_%s' % TaskStates.SLEEP)

    def awake(self):

        self.reload()
        if self.model.state != TaskStates.SLEEP:
            return

        self.model.update(state=TaskStates.PROCESSING)
        self._do_callback('TASK_STATE_%s' % TaskStates.PROCESSING)
        for t in self.model.descendants.filter(state=TaskStates.SLEEP):
            t.update(state=TaskStates.PROCESSING)
            SeaflowTask.get(task=t)._do_callback('TASK_STATE_%s' % TaskStates.PROCESSING)

        # steps
        for s in self.model.all_steps.prefetch_related('node').filter(state=TaskStates.SLEEP):
            if s.node.action_type == ActionTypes.Carrier:
                # 旁路节点
                func = tasks.start_carrier_step
            elif s.node.action_type == ActionTypes.External:
                # 外部节点
                func = tasks.publish_external_step
            else:
                # 普通节点
                func = get_func(s.node.action.func)
            func.delay(s.id)

    def _apply_dag(self, dag):
        """
        :param dag:子dag
        :return:
        """

        if not self._ready_to_execute_dag(dag):
            return
        self.seagull.info('apply dag 【%s】...' % dag.name)
        # 查找previous steps和previous tasks
        previous_tasks = Task.objects.filter(parent=self.model, dag__in=dag.previous_dags.all(),
                                             state=TaskStates.SUCCESS) \
            .exclude(iter_end=False)
        previous_steps = Step.objects.filter(task=self.model, node__in=dag.previous_nodes.all(),
                                             state=StepStates.SUCCESS) \
            .exclude(iter_end=False)
        if not previous_tasks and not previous_steps:
            # parent dag的第一批子dag
            inputs = self.model.input
        else:
            # inputs
            inputs = merge_outputs(merge_outputs_of_tasks(previous_tasks), merge_outputs_of_steps(previous_steps))

        self.seagull.info('dag 【%s】 raw input: %s'
                          % (dag.name, json.dumps(inputs, ensure_ascii=False)))
        input_adapter = ParamAdapter.from_json(dag.input_adapter)

        def _apply(_inputs,
                   _fissionable=False, _fission_index=0, _fission_count=1,
                   _iterable=False, _iter_index=0, _iter_context=None):

            extra = {}
            if _iter_context: extra['iter_context'] = _iter_context
            t, created = Task.objects.get_or_create(
                defaults=dict(
                    name=dag.name,
                    title=dag.title,
                    state=TaskStates.PROCESSING,
                    fission_count=_fission_count,
                    iter_end=False if _iterable else None,
                    input=_inputs,
                    config=self._generate_task_config(dag),
                    start_time=timezone.now(),
                    root=self.model.root or self.model,
                    extra=extra
                ),
                dag=dag,
                parent=self.model,
                fission_index=_fission_index,
                iter_index=_iter_index
            )
            s_task = SeaflowTask.get(task=t)
            if not created:
                self.seagull.warn('task 【%s】%s%s already exists'
                                  % (t.name,
                                     ' fission-%s' % _fission_index if _fissionable else '',
                                     ' iter-%s' % _iter_index if _iterable else ''
                                     ))
                self.seagull.flush(True)
                # 其他进程中已经创建了这个任务
                return
            t.previous_tasks.set(previous_tasks)
            t.previous_steps.set(previous_steps)

            try:
                s_task.seagull.info('task 【%s】%s%s created: %s'
                                    % (t.name,
                                       ' fission-%s' % _fission_index if _fissionable else '',
                                       ' iter-%s' % _iter_index if _iterable else '',
                                       t.id))

                # 第一批free node
                _nodes = dag.nodes.filter(previous_dags=None, previous_nodes=None)
                if _nodes:
                    s_task.seagull.info('apply nodes...')
                    [s_task._apply_node(n) for n in _nodes]

                # 第一批dag
                _dags = dag.children.filter(previous_dags=None, previous_nodes=None)
                if _dags:
                    s_task.seagull.info('apply dags...')
                    [s_task._apply_dag(d) for d in _dags]
                s_task.seagull.flush(True)
            except Exception as e:
                s_task.seagull.flush(True)
                s_task._break_off(e)

        if dag.fissionable:
            # 分裂
            inputs = fission_inputs(inputs, dag.fission_config['key'])
            fission_count = len(inputs)
            self.seagull.info('dag 【%s】 fission to %s by %s'
                              % (dag.name, fission_count, dag.fission_config['key']))
            for i, ii in enumerate(inputs):
                # adapt
                if dag.iterable:
                    ii, iter_key, iter_sequence = iter_inputs(ii, dag.iter_config['key'])
                    iter_index = 0
                    iter_context = {
                        'key': iter_key,
                        'sequence': iter_sequence
                    }
                    self.seagull.info('dag 【%s】 fission-%s start iter-%s'
                                      % (dag.name, i, iter_index))
                else:
                    iter_index = 0
                    iter_context = None
                ii = input_adapter.adapt(ii)
                self.seagull.info(
                    'dag 【%s】 fission-%s%s input: %s'
                    % (dag.name, i,
                       ' iter-%s' % iter_index if dag.iterable else '',
                       json.dumps(ii, ensure_ascii=False)))
                _apply(ii, True, i, fission_count, dag.iterable, iter_index, iter_context)
        else:
            if dag.iterable:
                inputs, iter_key, iter_sequence = iter_inputs(inputs, dag.iter_config['key'])
                iter_index = 0
                iter_context = {
                    'key': iter_key,
                    'sequence': iter_sequence
                }
                self.seagull.info('dag 【%s】 start iter-%s'
                                  % (dag.name, iter_index))
            else:
                iter_index = 0
                iter_context = None
            inputs = input_adapter.adapt(inputs)
            self.seagull.info('dag 【%s】%s input: %s'
                              % (dag.name,
                                 ' iter-%s' % iter_index if dag.iterable else '',
                                 json.dumps(inputs, ensure_ascii=False)))
            _apply(inputs, False, 0, 1, dag.iterable, iter_index, iter_context)

        self.seagull.flush(True)

    def _apply_node(self, node):
        """apply node, 请注意iter node的非首个step不会在这里处理
        :param node:
        :return:
        """
        ready = self._ready_to_execute_node(node)
        if not ready:
            return
        self.seagull.info('apply node 【%s】...' % node.name)
        # 查找previous steps和previous tasks
        previous_tasks = Task.objects.filter(parent=self.model,
                                             dag__in=node.previous_dags.all(),
                                             state=TaskStates.SUCCESS) \
            .exclude(iter_end=False)
        previous_steps = Step.objects.filter(task=self.model,
                                             node__in=node.previous_nodes.all(),
                                             state=StepStates.SUCCESS). \
            exclude(iter_end=False)
        # 参数
        if not previous_tasks and not previous_steps:
            # dag的第一批node
            inputs = self.model.input
        else:
            # inputs
            inputs = merge_outputs(merge_outputs_of_tasks(previous_tasks), merge_outputs_of_steps(previous_steps))

        self.seagull.info('node 【%s】 raw input: %s' % (node.name, json.dumps(inputs, ensure_ascii=False)))
        input_adapter = ParamAdapter.from_json(node.input_adapter)
        input_def = ParamDefinition.from_json(node.action.input_def)

        def _apply(_inputs,
                   _fissionable=False, _fission_index=0, _fission_count=1,
                   _iterable=False, _iter_index=0, _iter_context=None):

            extra = {}
            if _iter_context: extra['iter_context'] = _iter_context

            s, created = Step.objects.get_or_create(
                defaults=dict(
                    state=StepStates.PENDING,
                    fission_count=_fission_count,
                    iter_end=False if _iterable else None,
                    input=_inputs,
                    config=self._generate_step_config(node),
                    name=node.name,
                    title=node.title,
                    root=self.model.root or self.model,
                    extra=extra
                ),
                node=node,
                task=self.model,
                fission_index=_fission_index,
                iter_index=_iter_index
            )
            s_step = SeaflowStep.get(step=s)
            if not created:
                self.seagull.warn('step 【%s】%s%s already exists'
                                  % (s.name,
                                     ' fission-%s' % _fission_index if _fissionable else '',
                                     ' iter-%s' % _iter_index if _iterable else ''
                                     ))
                self.seagull.flush(True)
                # 其他进程中已经创建了这个任务
                return
            s.previous_tasks.set(previous_tasks)
            s.previous_steps.set(previous_steps)

            try:
                s_step.seagull.info('step 【%s】%s%s created: %s'
                                    % (s.name,
                                       ' fission-%s' % _fission_index if _fissionable else '',
                                       ' iter-%s' % _iter_index if _iterable else '',
                                       s.id))
                s_step.seagull.info('action type: %s' % node.action_type)
                if node.action_type == ActionTypes.Carrier:
                    # 旁路节点
                    func = tasks.start_carrier_step
                elif node.action_type == ActionTypes.External:
                    # 外部节点
                    func = tasks.publish_external_step
                else:
                    # 普通节点
                    func = get_func(node.action.func)

                s_step.seagull.flush(True)

                func.apply_async((s.id,), countdown=s.config.get('countdown', 0))
            except Exception as e:
                s_step.seagull.flush(True)
                s_step._break_off(e)

        if node.fissionable:
            # 分裂
            inputs = fission_inputs(inputs, node.fission_config['key'])
            fission_count = len(inputs)
            self.seagull.info('node 【%s】 fission to %s by %s'
                              % (node.name, fission_count, node.fission_config['key']))
            for i, ii in enumerate(inputs):
                if node.iterable:
                    ii, iter_key, iter_sequence = iter_inputs(ii, node.iter_config['key'])
                    iter_index = 0
                    iter_context = {
                        'key': iter_key,
                        'sequence': iter_sequence
                    }
                    self.seagull.info('node 【%s】 fission-%s start iter-%s'
                                      % (node.name, i, iter_index))
                else:
                    iter_index = 0
                    iter_context = None
                # adapt
                if not (node.action_type == ActionTypes.Carrier and not input_adapter):
                    # carrier特殊处理
                    ii = input_adapter.adapt(ii)
                if not (node.action_type == ActionTypes.Carrier and not input_def):
                    # carrier特殊处理
                    valid, errors = input_def.validate(ii)
                    if not valid:
                        self.seagull.error('node 【%s】 fission-%s%s validate '
                                           'input failed: expect %s, got %s'
                                           % (node.name,
                                              i,
                                              ' iter-%s' % iter_index if node.iterable else '',
                                              input_def.dump(),
                                              ii))
                        raise ParamDefinitionException(*errors.keys())

                self.seagull.info(
                    'node 【%s】 fission-%s%s input: %s'
                    % (node.name, i,
                       ' iter-%s' % iter_index if node.iterable else '',
                       json.dumps(ii, ensure_ascii=False)))
                _apply(ii, node.fissionable, i, fission_count, node.iterable, iter_index, iter_context)
        else:
            if node.iterable:
                inputs, iter_key, iter_sequence = iter_inputs(inputs, node.iter_config['key'])
                iter_index = 0
                iter_context = {
                    'key': iter_key,
                    'sequence': iter_sequence
                }
                self.seagull.info('node 【%s】 start iter-%s' % (node.name, iter_index))
            else:
                iter_index = 0
                iter_context = None
            if not (node.action_type == ActionTypes.Carrier and not input_adapter):
                # carrier特殊处理
                inputs = input_adapter.adapt(inputs)
            if not (node.action_type == ActionTypes.Carrier and not input_def):
                # carrier特殊处理
                valid, errors = input_def.validate(inputs)
                if not valid:
                    self.seagull.error('node 【%s】%s validate input failed: expect %s, got %s'
                                       % (node.name,
                                          ' iter-%s' % iter_index if node.iterable else '',
                                          input_def.dump(),
                                          inputs))
                    raise ParamDefinitionException(*errors.keys())

            self.seagull.info('node 【%s】%s input: %s'
                              % (node.name,
                                 ' iter-%s' % iter_index if node.iterable else '',
                                 json.dumps(inputs, ensure_ascii=False)))
            _apply(inputs, node.fissionable, 0, 1, node.iterable, iter_index, iter_context)
        self.seagull.flush(True)

    def _break_off(self, e=None, outputs={}):
        self.reload()
        if self.model.state not in [TaskStates.PENDING, TaskStates.PROCESSING, TaskStates.RETRY]:
            return
        seagull = Seagull.instance(self.model)

        if e:
            error = seagull.trace_error(e)
            if self.model.retries < self.model.config.get('max_retries', 0) and not isinstance(e, RevokeException):
                state = TaskStates.RETRY
            elif isinstance(e, TimeoutException):
                state = TaskStates.TIMEOUT
            elif isinstance(e, RevokeException):
                state = TaskStates.REVOKE
            else:
                state = TaskStates.ERROR
        else:
            seagull.info('output: %s' % json.dumps(outputs, cls=ComplexJSONEncoder))
            error = ""
            if self.model.retries < self.model.config.get('max_retries', 0):
                state = TaskStates.RETRY
            else:
                state = TaskStates.ERROR

        seagull.error('task 【%s】 broken: %s' % (self.model.name, state.value))
        if state == TaskStates.RETRY:
            seagull.info('retry in %ss.' % self.model.dag.retry_countdown)

        seagull.flush(True, merge=True)
        end_time = timezone.now()
        duration = end_time - self.model.start_time
        self.model.update(
            state=state,
            end_time=end_time,
            duration=float('%s.%s' % (duration.seconds, duration.microseconds)),
            error=error,
            output=outputs or {}
        )

        self._do_callback('TASK_STATE_%s' % self.model.state)
        # propagate
        if self.parent:
            self.parent._break_off(e, outputs=outputs)

    def _finish(self, outputs={}):
        self.reload()
        if self.model.state != TaskStates.PROCESSING:
            return

        self.seagull.info('task 【%s】%s%s finished: %s'
                          % (self.model.name,
                             ' fission-%s' % self.model.fission_index if self.model.dag.fissionable else '',
                             ' iter-%s' % self.model.iter_index if self.model.dag.iterable else '',
                             TaskStates.SUCCESS.value))
        self.seagull.info('output: %s' % json.dumps(outputs, cls=ComplexJSONEncoder))

        end_time = timezone.now()
        duration = end_time - self.model.start_time

        self.seagull.flush(True, merge=True)
        self.model.update(
            output=outputs,
            state=TaskStates.SUCCESS,
            end_time=end_time,
            duration=float('%s.%s' % (duration.seconds, duration.microseconds)),
            iter_end=self._is_iter_end()
        )

        self._do_callback('TASK_STATE_%s' % TaskStates.SUCCESS)

        if self.parent:
            # 路在何方
            self._forward()

    def _forward(self):
        try:
            outputs = self.model.output
            self.parent.seagull.info('dag 【%s】%s%s finished'
                                     % (self.model.dag.name,
                                        ' fission-%s' % self.model.fission_index if self.model.dag.fissionable else '',
                                        ' iter-%s' % self.model.iter_index if self.model.dag.iterable else '',
                                        ))
            if self.model.dag.iterable:
                if not self.model.iter_end:
                    # 继续iter
                    self.parent.seagull.info('dag 【%s】%s forward to iter-%s' % (self.model.dag.name,
                                                                                ' fission-%s' % self.model.fission_index
                                                                                if self.model.dag.fissionable else '',
                                                                                self.model.iter_index + 1,
                                                                                ))
                    self.parent._iter_dag(self.model.dag, self.model.fission_index, self.model.iter_index + 1)
                    return
            # 是否是fission
            if self.model.dag.fissionable:
                # 判断兄弟们是否完成
                siblings = Task.objects.filter(parent=self.model.parent, dag=self.model.dag).exclude(iter_end=False)
                if len([s for s in siblings if s.state == TaskStates.SUCCESS]) != self.model.fission_count:
                    self.parent.seagull.flush(True)
                    return
                self.parent.seagull.info('dag 【%s】 all fissions finished' % self.model.name)
                outputs = merge_fission_outputs(*[s.output for s in siblings])

            self.parent.seagull.info(
                'dag 【%s】 output: %s' % (self.model.dag.name, json.dumps(outputs, ensure_ascii=True)))

            # 有没有后继dag和node
            next_dags, next_nodes = self.model.dag.next_dags.all(), self.model.dag.next_nodes.all()
            if next_dags or next_nodes:
                if next_dags:
                    self.parent.seagull.info('apply next dags from 【%s】...' % self.model.dag.name)
                    [self.parent._apply_dag(d) for d in next_dags]
                if next_nodes:
                    self.parent.seagull.info('apply next nodes from 【%s】...' % self.model.dag.name)
                    [self.parent._apply_node(n) for n in next_nodes]
            else:
                # 自己处于尾部, 判断父task是否完成
                parent_finished, parent_outputs = self.parent._is_finished()
                if parent_finished:
                    self.parent._finish(self.parent._adapt_outputs(parent_outputs))
        except Exception as e:
            self.parent._break_off(e)
        finally:
            self.parent.seagull.flush(True)

    def _iter_dag(self, dag, fission_index, iter_index):
        """
        :param fission_index:
        :param iter_index:
        :param dag:
        :return:
        """
        # previous_task
        previous_task = Task.objects.get(parent=self.model, dag=dag, fission_index=fission_index,
                                         iter_index=iter_index - 1,
                                         state=TaskStates.SUCCESS)

        iter_index = previous_task.iter_index + 1
        inputs = previous_task.output
        iter_context = previous_task.extra['iter_context']

        if dag.iter_config.get('key'):
            # iter by key
            inputs[iter_context['key']] = iter_context['sequence'][iter_index]

        inputs = ParamAdapter.from_json(dag.input_adapter).adapt(inputs or {})
        t, created = Task.objects.get_or_create(
            defaults=dict(
                name=dag.name,
                title=dag.title,
                state=TaskStates.PROCESSING,
                fission_count=previous_task.fission_count,
                input=inputs,
                config=self._generate_task_config(dag),
                start_time=timezone.now(),
                root=self.model.root or self.model,
                extra={'iter_context': iter_context}
            ),
            dag=dag,
            parent=self.model,
            fission_index=fission_index,
            iter_index=iter_index
        )
        s_task = SeaflowTask.get(task=t)
        if not created:
            self.seagull.warn('task 【%s】%s%s already exists'
                              % (t.name,
                                 ' fission-%s' % fission_index
                                 if dag.fissionable else '',
                                 ' iter-%s' % iter_index if dag.iterable else ''
                                 ))
            # 其他进程中已经创建了这个任务, 理论上不可能
            self.seagull.flush(True)
            return
        t.previous_tasks.set([previous_task])

        try:
            s_task.seagull.info('task 【%s】%s%s created: %s'
                                % (t.name,
                                   ' fission-%s' % fission_index
                                   if dag.fissionable else '',
                                   ' iter-%s' % iter_index if dag.iterable else '',
                                   t.id))
            # 第一批free node
            _nodes = dag.nodes.filter(previous_dags=None, previous_nodes=None)
            if _nodes:
                s_task.seagull.info('apply nodes...')
                [s_task._apply_node(n) for n in _nodes]

            # 第一批dag
            _dags = dag.children.filter(previous_dags=None, previous_nodes=None)
            if _dags:
                s_task.seagull.info('apply dags...')
                [s_task._apply_dag(d) for d in _dags]
            s_task.seagull.flush(True)
        except Exception as e:
            s_task.seagull.flush(True)
            s_task._break_off(e)

    def _iter_node(self, node, fission_index, iter_index):
        """
        :param fission_index:
        :param iter_index:
        :param node:
        :return:
        """
        # previous_step
        previous_step = Step.objects.get(task=self.model, node=node, fission_index=fission_index,
                                         iter_index=iter_index - 1,
                                         state=StepStates.SUCCESS)

        iter_index = previous_step.iter_index + 1
        inputs = previous_step.output
        iter_context = previous_step.extra['iter_context']

        if node.iter_config.get('key'):
            # iter by key
            inputs[iter_context['key']] = iter_context['sequence'][iter_index]

        inputs = ParamAdapter.from_json(node.input_adapter).adapt(inputs or {})
        s, created = Step.objects.get_or_create(
            defaults=dict(
                state=StepStates.PENDING,
                fission_count=previous_step.fission_count,
                input=inputs,
                config=self._generate_step_config(node),
                name=node.name,
                title=node.title,
                root=self.model.root or self.model,
                extra={'iter_context': iter_context}
            ),
            node=node,
            task=self.model,
            fission_index=fission_index,
            iter_index=iter_index
        )
        s_step = SeaflowStep.get(step=s)
        if not created:
            self.seagull.warn('step 【%s】%s%s already exists'
                              % (s.name,
                                 ' fission-%s' % fission_index
                                 if node.fissionable else '',
                                 ' iter-%s' % iter_index if node.iterable else ''
                                 ))
            # 其他进程中已经创建了这个任务, 理论上不可能
            self.seagull.flush(True)
            return
        s.previous_steps.set([previous_step])

        try:
            s_step.seagull.info('step 【%s】%s%s created: %s'
                                % (s.name,
                                   ' fission-%s' % previous_step.fission_index
                                   if node.fissionable else '',
                                   ' iter-%s' % iter_index if node.iterable else '',
                                   s.id))
            s_step.seagull.info('action type: %s' % node.action_type)
            if node.action_type == ActionTypes.External:
                # 外部节点
                func = Seaflow.publish_external_step
            elif node.action_type == ActionTypes.Carrier:
                # 旁路节点
                func = tasks.start_carrier_step
            else:
                func = get_func(node.action.func)

            s_step.seagull.flush(True)
            func.delay(s.id)
        except Exception as e:
            s_step.seagull.flush(True)
            s_step._break_off(e)

    def _adapt_outputs(self, outputs):
        # outputs = outputs or {}

        output_adapter = ParamAdapter.from_json(self.model.dag.output_adapter)
        outputs = output_adapter.adapt(outputs)

        return outputs

    def _is_iter_end(self):
        if not self.model.dag.iterable:
            return
        iter_context = self.model.extra['iter_context']
        if self.model.dag.iter_config.get('key'):
            return self.model.iter_index == (len(iter_context['sequence']) - 1)

    def _ready_to_execute_node(self, node):
        """
        ready to execute node in task
        :param node:
        :return:
        """
        # previous nodes
        for n in node.previous_nodes.all():
            rs = Step.objects.filter(task=self.model, node=n, state=StepStates.SUCCESS) \
                .exclude(iter_end=False)
            if not rs:
                # previous node还么有已完成的step
                return False
            if len(rs) != rs[0].fission_count:
                # previous node完成的step数没有达到额定分裂数量
                return False
        # previous dags
        for d in node.previous_dags.all():
            rs = Task.objects.filter(parent=self.model, dag=d, state=TaskStates.SUCCESS.name).exclude(iter_end=False)
            if not rs:
                # previous dag还么有已完成的task
                return False
            if len(rs) != rs[0].fission_count:
                # previous dag完成的task数没有达到额定分裂数量
                return False
        return True

    def _ready_to_execute_dag(self, dag):
        """
        ready to execute dag in task
        :param dag:
        :return:
        """

        # previous nodes
        for n in dag.previous_nodes.all():
            rs = Step.objects.filter(task=self.model, node=n, state=StepStates.SUCCESS) \
                .exclude(iter_end=False)
            if not rs:
                # previous node还么有已完成的step
                return False
            if len(rs) != rs[0].fission_count:
                # previous node完成的step数没有达到额定分裂数量
                return False
        # previous dags
        for d in dag.previous_dags.all():
            rs = Task.objects.filter(parent=self.model, dag=d, state=TaskStates.SUCCESS).exclude(iter_end=False)
            if not rs:
                # previous dag还么有已完成的task
                return False
            if len(rs) != rs[0].fission_count:
                # previous dag完成的task数没有达到额定分裂数量
                return False
        return True

    def _is_dag_finished(self, dag):
        """
        判断某个子dag是否执行完毕: dag对应的task全部完成并达到额定分裂数量
        :param dag:
        :return:
            finished: bool
            outputs: dict
        """
        rs = Task.objects.filter(parent=self.model, dag=dag, state=TaskStates.SUCCESS) \
            .exclude(iter_end=False)
        if not rs:
            # 还没有已完成的task
            return False, None
        if len(rs) != rs[0].fission_count:
            # 已完成的task数没有达到额定分裂数量
            return False, None

        if dag.fissionable:
            outputs = merge_fission_outputs(*[r.output for r in rs])
        else:
            outputs = rs[0].output

        return True, outputs

    def _is_node_finished(self, node):
        """
        判断某个node是否执行完毕
        :param node:
        :return:
        """
        rs = Step.objects.filter(task=self.model, node=node,
                                 state=StepStates.SUCCESS) \
            .exclude(iter_end=False)
        if not rs:
            # 还么有已完成的step
            return False, None
        if len(rs) != rs[0].fission_count:
            # 已完成的step数没有达到额定分裂数量
            return False, None

        if node.fissionable:
            outputs = merge_fission_outputs(*[r.output for r in rs])
        else:
            outputs = rs[0].output

        return True, outputs

    def _is_finished(self):
        """
        dag中所有的尾结点(dag/node)都执行完毕
        :return:
            finished: bool
            outputs: dict
        """
        outputs_list = []
        for dag in self.model.dag.children.filter(next_dags=None, next_nodes=None):
            finished, outputs = self._is_dag_finished(dag)
            if not finished:
                return False, None
            outputs_list.append(outputs)
        for node in self.model.dag.nodes.filter(next_dags=None, next_nodes=None):
            finished, outputs = self._is_node_finished(node)
            if not finished:
                return False, None
            outputs_list.append(outputs)

        outputs = merge_outputs(*outputs_list)

        return True, outputs

    def _is_loop_end(self, outputs):
        """
        判断下一次loop是否满足loop_condition
        :param outputs:
        :return:
        """
        data = {
            'index': self.model.loop_index + 1,
            'input': self.model.input,
            'output': outputs,
            'context': (self.model.root or self.model).context
        }

        loop_condition = self.model.dag.loop_config['condition']

        return not jsonLogic(loop_condition, data)

    def _generate_task_config(self, dag):
        """
        :param dag: sub dag
        :return:
        """
        task_config = TaskConfig.from_json(
            (self.root or self).model.extra.get('tasks_config', {}).get(dag.name, {})).trim()
        config = TaskConfig(
            max_retries=dag.max_retries,
            retry_countdown=dag.retry_countdown,
            timeout=dag.timeout,
        ).trim()
        config.update(task_config)
        return config

    def _generate_step_config(self, node):
        """
        :param node:
        :return:
        """
        step_config = TaskConfig.from_json(
            (self.root or self).model.extra.get('steps_config', {}).get(node.name, {})).trim()
        config = StepConfig(
            max_retries=node.max_retries,
            retry_countdown=node.retry_countdown,
            timeout=node.timeout,
        ).trim()
        config.update(step_config)
        return config

    def _do_callback(self, event):
        """
        :param event:
        :return:
        """
        if self.model.config.get('callback'):
            cb = self.model.config.get('callback')
            if cb['is_async']:
                do_callback.apply_async((cb['func'], event, self.model.to_json()))
            else:
                do_callback.apply((cb['func'], event, self.model.to_json()))


class SeaflowStep(object):

    def __init__(self):
        # model
        self.model = None
        self.id = None
        self.name = None
        self.context = None
        self.seagull = None
        self.config = None

        #
        self.task = None  # SeaflowTask
        self.root = None  # SeaflowTask

    @classmethod
    def get(cls, step_id=None, step=None):
        """
        :param step_id:
        :param step:
        :return: SeaflowStep
        """
        m = step or Step.objects.select_related('node', 'node__action',
                                                'task', 'root').defer('logs').get(pk=step_id)
        r = cls()
        r.model = m
        r.load()
        return r

    def load(self):
        self.id = self.model.id
        self.name = self.model.name
        self.context = self.model.root.context
        self.seagull = Seagull.instance(self.model)
        self.config = self.model.config

        self.task = SeaflowTask.get(task=self.model.task)
        self.root = SeaflowTask.get(task=self.model.root)

    def reload(self):
        self.model.refresh_from_db()
        self.load()

    def _dispatch(self, identity={}):
        """
        派发外部任务
        :param identity: dict，领取者的身份信息
                    name:
                    key:
        :return: identifier
        """
        if self.model.state != StepStates.PUBLISH:
            raise StateException('step is not in publish state')
        if self.model.node.action_type != ActionTypes.External:
            raise SeaflowException('step is not external')
        try:
            identifier = generate_identifier('%s-%s' % (self.id, identity.get('key')), rand=False)
            self.model.update(identifier=identifier, start_time=timezone.now(),
                              state=StepStates.PROCESSING)
            self._do_callback('STEP_STATE_%s' % StepStates.PROCESSING)
            self.seagull.info('step subscribed by %s' % (identity.get('name')))
            self.seagull.info('step execution...')
        finally:
            self.seagull.flush(True)

        return identifier

    def _execute(self, celery_action):
        celery_action.step = self.model
        celery_action.seagull = self.seagull
        try:
            celery_action.task = self.model.task
            celery_action.root = self.model.root
            celery_action.context = SeaflowContext(task=self.model.root)

            if self.model.state == StepStates.PENDING:
                # fresh new
                self.seagull.info('step 【%s】 started' % self.name)
                self.seagull.info('input: %s' % json.dumps(self.model.input, ensure_ascii=False))
                if self.model.node.loopable:
                    self.seagull.info('loop started')
                    self.seagull.info('loop-%s started' % self.model.loop_index)
            # 是否睡眠状态
            elif self.model.state == StepStates.SLEEP:
                self.seagull.info('Good Morning.')
            elif self.model.state == StepStates.PROCESSING:
                if self.model.node.loopable:
                    self.seagull.info('loop-%s started' % self.model.loop_index)

            # 判断是否进入睡眠
            sleep = False
            if self.model.task.state == TaskStates.SLEEP:
                sleep = True
                self.seagull.info('detect task 【%s】 state: %s' % (self.model.task.name, self.model.task.state))
            elif self.model.root.state == TaskStates.SLEEP:
                sleep = True
                self.seagull.info(
                    'detect root task 【%s】 state: %s' % (self.model.root.name, self.model.root.state))
            if sleep:
                self.model.update(identifier=celery_action.request.id, state=StepStates.SLEEP)
                self._do_callback('STEP_STATE_%s' % StepStates.SLEEP)
                self.seagull.info('Good Night.')
                self.seagull.flush(True)
                return

            if self.model.state in [StepStates.PENDING, StepStates.RETRY, StepStates.SLEEP]:
                # fresh new/retry/awake from sleep
                self.model.update(identifier=celery_action.request.id, start_time=timezone.now(),
                                  state=StepStates.PROCESSING)
                self._do_callback('STEP_STATE_%s' % StepStates.PROCESSING)
            else:
                # loop/recovery
                self.model.update(identifier=celery_action.request.id)

            # 是否撤销
            if self.model.task.state in [
                TaskStates.ERROR,
                TaskStates.TIMEOUT,
                TaskStates.REVOKE,
                TaskStates.TERMINATE,
            ]:
                raise RevokeException(
                    'detect task 【%s】 state: %s' % (self.model.task.name, self.model.task.state))
            if self.model.root.state in [
                TaskStates.ERROR,
                TaskStates.TIMEOUT,
                TaskStates.REVOKE,
                TaskStates.TERMINATE,
            ]:
                raise RevokeException(
                    'detect root task 【%s】 state: %s' % (self.model.root.name, self.model.root.state))

            self.seagull.info('step execution...')
            if self.model.config.get('timeout'):
                res = timeout(self.model.config.get('timeout'),
                              timeout_exception=TimeoutException)(celery_action.func)(celery_action, **self.model.input)
            else:
                res = celery_action.func(celery_action, **self.model.input)
            res = res or {}
            state, outputs = res.get('state', StepStates.SUCCESS), res.get('data', {})
            # self.seagull.info('action【%s】 output: %s' % (self.model.node.action.name,
            #                                              json.dumps(outputs, cls=ComplexJSONEncoder)))
            self.seagull.flush(True)
            outputs = self._adapt_outputs(outputs)
            if state == StepStates.SUCCESS:
                if self.model.node.loopable:
                    # loop step
                    self._loop_next(outputs)
                else:
                    # finish step
                    self._finish(outputs, state)
            else:
                self._break_off(outputs=outputs)
        except Exception as e:
            self._break_off(e)
        finally:
            del_attrs(celery_action, 'seagull', 'task', 'root_task', 'step', 'context', 'func')

    def _publish(self):
        try:
            # 是否睡眠状态
            if self.model.state == StepStates.SLEEP:
                # 唤醒
                self.seagull.info('Good Morning.')
            else:
                self.seagull.info('step 【%s】 started' % self.name)
                self.seagull.info('input: %s' % json.dumps(self.model.input, ensure_ascii=False))
                # 判断是否进入睡眠
                sleep = False
                if self.model.task.state == TaskStates.SLEEP:
                    sleep = True
                    self.seagull.info('detect task 【%s】 state: %s' % (self.model.task.name, self.model.task.state))
                elif self.model.root.state == TaskStates.SLEEP:
                    sleep = True
                    self.seagull.info(
                        'detect root task 【%s】 state: %s' % (self.model.root.name, self.model.root.state))
                if sleep:
                    self.model.update(state=StepStates.SLEEP)
                    self._do_callback('STEP_STATE_%s' % StepStates.SLEEP)
                    self.seagull.info('Good Night.')
                    self.seagull.flush(True)
                    return

            # 是否撤销
            if self.model.task.state in [
                TaskStates.ERROR,
                TaskStates.TIMEOUT,
                TaskStates.REVOKE,
                TaskStates.TERMINATE,
            ]:
                raise RevokeException(
                    'detect task 【%s】 state: %s' % (self.model.task.name, self.model.task.state))
            if self.model.root.state in [
                TaskStates.ERROR,
                TaskStates.TIMEOUT,
                TaskStates.REVOKE,
                TaskStates.TERMINATE,
            ]:
                raise RevokeException(
                    'detect root task 【%s】 state: %s' % (self.model.root.name, self.model.root.state))

            self.model.update(state=StepStates.PUBLISH)
            self._do_callback('STEP_STATE_%s' % StepStates.PUBLISH)
            self.seagull.info('step published')
            self.seagull.flush(True)
        except Exception as e:
            self._break_off(e)

    def _carry(self, celery_action):
        try:
            self.model.update(identifier=celery_action.request.id, start_time=timezone.now(),
                              state=StepStates.PROCESSING)
            self._do_callback('STEP_STATE_%s' % StepStates.PROCESSING)
            self.seagull.info('step 【%s】 started' % self.name)
            # 是否撤销
            revoke = self.model.task.state in [
                TaskStates.ERROR,
                TaskStates.TIMEOUT,
                TaskStates.REVOKE,
                TaskStates.TERMINATE,
            ] or self.model.root.state in [
                         TaskStates.ERROR,
                         TaskStates.TIMEOUT,
                         TaskStates.REVOKE,
                         TaskStates.TERMINATE,
                     ]

            if revoke:
                raise RevokeException(
                    'detect task 【%s】 state: %s' % (self.model.task.name, self.model.task.state))

            self.seagull.flush(True)
            self._finish(self._adapt_outputs(self.model.input))
        except Exception as e:
            self._break_off(e)

    def _finish(self, outputs={}, state=StepStates.SUCCESS):
        self.reload()
        if self.model.state != StepStates.PROCESSING:
            return

        self.seagull.info('step 【%s】%s%s finished: %s'
                          % (self.name,
                             ' fission-%s' % self.model.fission_index if self.model.node.fissionable else '',
                             ' iter-%s' % self.model.iter_index if self.model.node.iterable else '',
                             state.value))
        self.seagull.info('output: %s' % json.dumps(outputs, cls=ComplexJSONEncoder))

        end_time = timezone.now()
        duration = end_time - self.model.start_time

        self.seagull.flush(True, merge=True)
        self.model.update(
            state=state,
            end_time=end_time,
            duration=float('%s.%s' % (duration.seconds, duration.microseconds)),
            output=outputs or {},
            iter_end=self._is_iter_end()
        )
        # 路在脚下
        self._forward()

    def _break_off(self, e=None, outputs={}):
        self.reload()
        if self.model.state not in [StepStates.PENDING, StepStates.PUBLISH, StepStates.PROCESSING, StepStates.RETRY,
                                    StepStates.SLEEP]:
            return

        if e:
            error = self.seagull.trace_error(e)
            if self.model.retries < self.model.config.get('max_retries', 0) and not isinstance(e, RevokeException):
                state = StepStates.RETRY
            elif isinstance(e, TimeoutException):
                state = StepStates.TIMEOUT
            elif isinstance(e, RevokeException):
                state = StepStates.REVOKE
            else:
                state = StepStates.ERROR
        else:
            self.seagull.info('output: %s' % json.dumps(outputs, cls=ComplexJSONEncoder))
            error = ""
            if self.model.retries < self.model.config.get('max_retries', 0):
                state = StepStates.RETRY
            else:
                state = StepStates.ERROR

        self.seagull.error('step 【%s】 broken: %s' % (self.name, state.value))
        if state == StepStates.RETRY:
            self.seagull.info('retry in %ss.' % self.model.node.retry_countdown)
        self.seagull.flush(True, merge=True)

        end_time = timezone.now()
        duration = (end_time - self.model.start_time) if self.model.start_time else None
        self.model.update(
            state=state,
            retries=self.model.retries + 1 if state == StepStates.RETRY else self.model.retries,
            end_time=None if state == StepStates.RETRY else end_time,
            duration=None if (not duration or state == StepStates.RETRY) else float(
                '%s.%s' % (duration.seconds, duration.microseconds)),
            error=error,
            output=outputs or {}
        )
        if state == StepStates.RETRY:
            # celery_task.retry(countdown=step.node.retry_countdown, exc=e, throw=False)
            get_func(self.model.node.action.func).apply_async((self.id,),
                                                              countdown=self.model.config.get('retry_countdown', 0))
        else:
            # propagate
            self.task._break_off(e, outputs=outputs)

    def _loop_next(self, outputs={}):
        self.reload()
        if self.model.state != StepStates.PROCESSING:
            return

        self.seagull.info(
            'loop-%s end, output: %s' % (self.model.loop_index, json.dumps(outputs, cls=ComplexJSONEncoder)))
        loop_end = self._is_loop_end(outputs)
        if loop_end:
            self.seagull.info('loop end')
            self._finish(outputs)
        else:
            self.model.update(loop_index=self.model.loop_index + 1, _refresh=False)
            countdown = self.model.node.loop_config.get('countdown', 0)
            self.seagull.info('sleep %ss...' % countdown)
            get_func(self.model.node.action.func).apply_async((self.id,), countdown=countdown)

    def _forward(self):
        try:
            outputs = self.model.output
            self.task.seagull.info('node 【%s】%s%s finished'
                                   % (self.model.node.name,
                                      ' fission-%s' % self.model.fission_index if self.model.node.fissionable else '',
                                      ' iter-%s' % self.model.iter_index if self.model.node.iterable else '',
                                      ))
            if self.model.node.iterable:
                if not self.model.iter_end:
                    # 继续iter
                    self.task.seagull.info('node 【%s】%s forward to iter-%s'
                                           % (self.model.node.name,
                                              ' fission-%s' % self.model.fission_index if self.model.node.fissionable else '',
                                              self.model.iter_index + 1,
                                              ))
                    self.task._iter_node(self.model.node, self.model.fission_index, self.model.iter_index + 1)
                    return

            # 是否是fission
            if self.model.node.fissionable:
                # 判断兄弟们是否完成
                siblings = Step.objects.filter(task=self.model.task, node=self.model.node).exclude(
                    iter_end=False)
                if len([s for s in siblings if
                        s.state == StepStates.SUCCESS]) != self.model.fission_count:
                    self.task.seagull.flush(True)
                    return
                self.task.seagull.info('node 【%s】 all fissions finished' % self.model.node.name)
                outputs = merge_fission_outputs(*[s.output for s in siblings])

            self.task.seagull.info(
                'node【%s】 output: %s' % (self.model.node.name, json.dumps(outputs, ensure_ascii=True)))

            # 有没有后继dag和node
            next_dags, next_nodes = self.model.node.next_dags.all(), self.model.node.next_nodes.all()
            if next_dags or next_nodes:
                if next_dags:
                    self.task.seagull.info('apply next dags from 【%s】...' % self.model.node.name)
                    [self.task._apply_dag(d) for d in next_dags]
                if next_nodes:
                    self.task.seagull.info('apply next nodes from 【%s】...' % self.model.node.name)
                    [self.task._apply_node(n) for n in next_nodes]
            else:
                # 自己处于尾部，判断所在task是否完成
                task_finished, task_outputs = self.task._is_finished()
                if task_finished:
                    self.task._finish(task_outputs)
        except Exception as e:
            self.task._break__off(e)
        finally:
            self.task.seagull.flush(True)

    def _adapt_outputs(self, outputs):
        # outputs = outputs or {}

        output_def = ParamDefinition.from_json(self.model.node.action.output_def)
        if not (self.model.node.action_type == ActionTypes.Carrier and not output_def):
            # carrier特殊处理
            valid, errors = output_def.validate(outputs)
            if not valid:
                self.seagull.error('validate output failed: expect %s, got %s'
                                   % (output_def.dump(),
                                      json.dumps(outputs,
                                                 cls=ComplexJSONEncoder,
                                                 ensure_ascii=False)))
                self.seagull.flush(True)
                raise ParamDefinitionException(*errors.keys())
        output_adapter = ParamAdapter.from_json(self.model.node.output_adapter)
        if not (self.model.node.action_type == ActionTypes.Carrier and not output_adapter):
            outputs = output_adapter.adapt(outputs)

        return outputs

    def _is_iter_end(self):
        if not self.model.node.iterable:
            return
        iter_context = self.model.extra['iter_context']
        if self.model.node.iter_config.get('key'):
            return self.model.iter_index == (len(iter_context['sequence']) - 1)

    def _is_loop_end(self, outputs):
        """
        判断下一次次loop是否满足loop_condition
        :param outputs:
        :return:
        """

        data = {
            'index': self.model.loop_index + 1,
            'input': self.model.input,
            'output': outputs,
            'context': (self.model.root or self.model.task).context
        }

        loop_condition = self.model.node.loop_config['condition']

        return not jsonLogic(loop_condition, data)

    def _do_callback(self, event):
        """
        :param event:
        :return:
        """
        if self.model.config.get('callback'):
            cb = self.model.config.get('callback')
            if cb['is_async']:
                do_callback.apply_async((cb['func'], event, self.model.to_json()))
            else:
                do_callback.apply((cb['func'], event, self.model.to_json()))



from django.apps import AppConfig
from django.utils.translation import gettext_lazy as _

__all__ = ['SeaflowConfig']

class SeaflowConfig(AppConfig):
    name = 'seaflow'
    verbose_name = _("Seaflow")



import datetime
import json
import logging
import os
import sys
import time
import traceback

from .models import Log, Task
from .utils import NotPrintException


class Tracker(object):
    """
    logger
    """

    def __init__(self,
                 name='',
                 level=logging.NOTSET,
                 max_buffer_lines=1,
                 auto_flush=True,
                 flush_action=None
                 ):
        self.logger = logging.getLogger(name)
        self.logger.level = level
        self._buf = []
        self.level = level
        self.max_buffer_lines = max_buffer_lines
        self.auto_flush = auto_flush
        self.flush_action = flush_action
        self.ignore_empty_lines = True

    def length(self):
        return len(self._buf)

    def debug(self, message):
        if self.level <= logging.DEBUG:
            self.write(self._format(message, 'DEBUG'))
            self.logger.debug(message)

    def info(self, message):
        if self.level <= logging.INFO:
            self.write(self._format(message, 'INFO'))
            self.logger.info(message)

    def warn(self, message):
        if self.level <= logging.WARNING:
            self.write(self._format(message, 'WARN'))
            self.logger.warning(message)

    def error(self, message):
        if self.level <= logging.ERROR:
            self.write(self._format(message, 'ERROR'))
            self.logger.error(message)

    def flush(self, exec_flush_action=False, **kwargs):
        logs = self._buf.copy()
        self._buf = []
        if exec_flush_action and self.flush_action:
            self.flush_action(logs, **kwargs)
        return logs

    def write(self, content):
        self._buf.append({'ts': time.time() * 1000, 'content': content})
        if self.auto_flush and len(self._buf) >= self.max_buffer_lines + 1:
            self.flush(exec_flush_action=True)

    def _format(self, content, level):
        content = content.strip('\n')
        # co_filename, func_lineno, co_name = self._find_caller()
        # return '[%s] %s %s:%s: [line:%s] %s' % (
        return '[%s] %s: %s' % (
            datetime.datetime.utcnow().strftime('%Y-%m-%d %H:%M:%S.%f')[:-4],
            level,
            # co_filename,
            # co_name,
            # func_lineno,
            content
        )

    def trace_error(self, e=None):
        if e:
            e_type, e_value, traceback_obj = type(e), e, e.__traceback__
        else:
            e_type, e_value, traceback_obj = sys.exc_info()[:3]
        self.logger.exception(e_value)

        lines = ['Type: %s' % e_type, 'Value: %s' % e_value]
        for line in traceback.format_exception(e_type, e_value, traceback_obj)[1:]:
            line = line.rstrip('\n')
            lines.append(line)
        if not isinstance(e, NotPrintException):
            self.error('ErrorStack:')
            for line in lines:
                self.write(line)

        return '\n'.join(lines)

    def _find_caller(self):
        """
        获取调用者信息, 用于记录file func lineno
        :return:
        """

        f = logging.currentframe()

        if f is not None:
            f = f.f_back
        rv = "(unknown file)", 0, "(unknown function)"
        while hasattr(f, "f_code"):
            co = f.f_code
            filename = os.path.normcase(co.co_filename)
            if filename == __file__:
                f = f.f_back
                continue
            rv = (co.co_filename, f.f_lineno, co.co_name)
            break
        return rv


class Seagull(Tracker):
    seagulls = {}

    def __init__(self, ref, *args, **kwargs):
        self.ref = ref
        self.ref_type = 'TASK' if isinstance(self.ref, Task) else 'STEP'
        kwargs['flush_action'] = self.persist
        kwargs['auto_flush'] = True
        super().__init__('seagull', *args, **kwargs)

    def persist(self, logs, merge=False):
        if logs:
            Log.objects.bulk_create([Log(
                ref_type=self.ref_type,
                ref_id=self.ref.id,
                ts=lg['ts'],
                content=lg['content']
            ) for lg in logs])
        if merge:
            self.merge()

    def merge(self):
        logs = []
        ids = []
        for lg in Log.objects.filter(ref_id=self.ref.id, ref_type=self.ref_type):
            logs.append({'ts': lg.ts, 'content': lg.content})
            ids.append(lg.id)
        if logs:
            self.ref.refresh_from_db()
            self.ref.update(_refresh=False, logs=(self.ref.logs + logs))

            # NOTE: 避免一次删除太多数据，引发 DB 报警：每次删除 500 条日志
            cursor = 0
            size = 500
            while cursor < len(ids):
                Log.objects.filter(pk__in=ids[cursor:cursor+size]).delete()
                cursor += size

    @classmethod
    def instance(cls, target, *args, **kwargs):
        identifier = (target.__class__.name, target.id)
        if identifier not in cls.seagulls:
            cls.seagulls[identifier] = cls(target, *args, **kwargs)
        return cls.seagulls[identifier]

    def __del__(self):
        self.flush(True)



from django.db import models

# Create your models here.
from .consts import TaskStates, StepStates, ActionTypes


class BaseModel(models.Model):

    def update(self, _refresh=True, **kwargs):
        self.__class__.objects.filter(pk=self.pk).update(**kwargs)
        if _refresh:
            self.refresh_from_db()

    class Meta:
        abstract = True


class Action(BaseModel):
    """
    流程Action
    a1，a2，a3，a4，a5，a6
    """

    id = models.AutoField(primary_key=True)

    name = models.CharField(max_length=64, db_index=True, unique=True)
    title = models.CharField(max_length=64)
    type = models.CharField(max_length=32, choices=[(item.name, item.value) for item in ActionTypes])
    func = models.CharField(max_length=128, unique=True, null=True)
    input_def = models.JSONField(default=dict)
    output_def = models.JSONField(default=dict)

    create_time = models.DateTimeField(auto_now_add=True)
    update_time = models.DateTimeField(auto_now=True)

    class Meta:
        managed = True
        db_table = 'seaflow_action'
        verbose_name = '流程Action'
        verbose_name_plural = verbose_name


class Dag(BaseModel):
    """
    流程图
      -C-
    A--B--E--F
      -D-
    """

    id = models.AutoField(primary_key=True)
    name = models.CharField(max_length=64, db_index=True)
    title = models.CharField(max_length=64)
    version = models.PositiveIntegerField(null=True, db_index=True)  # root dag才有version
    latest = models.BooleanField(null=True)

    root = models.ForeignKey('Dag', db_constraint=False, null=True,
                             related_name='descendants', on_delete=models.CASCADE)
    parent = models.ForeignKey('Dag', db_constraint=False, null=True,
                               related_name='children', on_delete=models.CASCADE)
    previous_dags = models.ManyToManyField('Dag', db_constraint=False, related_name='next_dags',
                                           db_table='seaflow_dag_to_dag')
    previous_nodes = models.ManyToManyField('Node', db_constraint=False, related_name='next_dags',
                                            db_table='seaflow_node_to_dag')

    input_adapter = models.JSONField(default=dict)
    output_adapter = models.JSONField(default=dict)

    # 分裂
    fissionable = models.BooleanField('可分裂', default=False)
    fission_config = models.JSONField(default=dict)

    # 迭代
    iterable = models.BooleanField('可迭代', default=False)
    iter_config = models.JSONField(default=dict)  # condition, iter_key

    # 循环
    loopable = models.BooleanField('可循环', default=False)
    loop_config = models.JSONField(default=dict)  # condition, countdown

    max_retries = models.IntegerField(default=0)
    retry_countdown = models.IntegerField(default=0)
    timeout = models.IntegerField(null=True)

    create_time = models.DateTimeField(auto_now_add=True)
    update_time = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.name

    class Meta:
        managed = True
        db_table = 'seaflow_dag'
        verbose_name = '流程图'
        verbose_name_plural = verbose_name


class Node(BaseModel):
    """
    流程节点
    n1，n2，n3，n4，n5，n6
    """

    id = models.AutoField(primary_key=True)

    name = models.CharField(max_length=64)
    title = models.CharField(max_length=64)
    action = models.ForeignKey('Action', db_constraint=False, on_delete=models.PROTECT)
    dag = models.ForeignKey('Dag', db_constraint=False, related_name='nodes', on_delete=models.CASCADE)
    root_dag = models.ForeignKey('Dag', db_constraint=False, related_name='all_nodes', on_delete=models.CASCADE)

    previous_nodes = models.ManyToManyField('Node', db_constraint=False, related_name='next_nodes',
                                            db_table='seaflow_node_to_node')
    previous_dags = models.ManyToManyField('Dag', db_constraint=False, related_name='next_nodes',
                                           db_table='seaflow_dag_to_node')

    input_adapter = models.JSONField(default=dict)
    output_adapter = models.JSONField(default=dict)

    action_type = models.CharField(max_length=32,
                                   choices=[(item.name, item.value) for item in ActionTypes])

    # 分裂
    fissionable = models.BooleanField('可分裂', default=False)
    fission_config = models.JSONField(default=dict)

    # 迭代
    iterable = models.BooleanField('可迭代', default=False)
    iter_config = models.JSONField(default=dict)  # condition, iter_key

    # 循环
    loopable = models.BooleanField('可循环', default=False)
    loop_config = models.JSONField(default=dict)  # condition, countdown

    max_retries = models.IntegerField(default=0)
    retry_countdown = models.IntegerField(default=0)
    timeout = models.IntegerField(null=True)

    create_time = models.DateTimeField(auto_now_add=True)
    update_time = models.DateTimeField(auto_now=True)

    def __str__(self):
        return self.name

    class Meta:
        managed = True
        db_table = 'seaflow_node'
        verbose_name = '流程节点'
        verbose_name_plural = verbose_name


class Task(BaseModel):
    """
    任务
    """

    id = models.AutoField(primary_key=True)

    name = models.CharField(max_length=64)
    title = models.CharField(max_length=64)
    dag = models.ForeignKey('Dag', db_constraint=False, on_delete=models.PROTECT)
    state = models.CharField(max_length=64,
                             choices=[(item.name, item.value) for item in TaskStates])

    root = models.ForeignKey('Task', db_constraint=False, null=True,
                             related_name='descendants', on_delete=models.CASCADE)
    parent = models.ForeignKey('Task', db_constraint=False, null=True,
                               related_name='children', on_delete=models.CASCADE)
    previous_tasks = models.ManyToManyField('Task', db_constraint=False, related_name='next_tasks',
                                            db_table='seaflow_task_to_task')

    previous_steps = models.ManyToManyField('Step', db_constraint=False, related_name='next_tasks',
                                            db_table='seaflow_step_to_task')
    config = models.JSONField('配置', default=dict)
    retries = models.IntegerField('重试次数', default=0)
    # 分裂
    fission_count = models.IntegerField('分裂数', default=1)
    fission_index = models.IntegerField('分裂序号', db_index=True, default=0)

    # 迭代
    iter_index = models.IntegerField('迭代序号', db_index=True, default=0)
    iter_end = models.BooleanField('迭代末端', null=True)

    # 循环
    loop_index = models.IntegerField('循环序号', default=0)

    input = models.JSONField('输入')
    context = models.JSONField('上下文', default=dict)
    output = models.JSONField('输出', default=dict)

    extra = models.JSONField(default=dict)

    error = models.TextField('错误信息', default='')
    logs = models.JSONField('执行日志', default=list)
    logfile = models.CharField('s3日志文件路径', max_length=256, default='')
    log_refetch_time = models.DateTimeField(null=True, db_index=True)

    start_time = models.DateTimeField(null=True)
    end_time = models.DateTimeField(null=True)
    duration = models.FloatField(null=True)

    create_time = models.DateTimeField(auto_now_add=True)
    update_time = models.DateTimeField(auto_now=True)

    def __str__(self):
        return '%s%s%s' % (
            self.name,
            ' fission-%s' % self.fission_index if self.dag.fissionable else '',
            ' iter-%s' % self.iter_index if self.dag.iterable else '',
        )

    def to_dict(self, with_logs=False):
        d = {
            'id': self.id,
            'name': self.name,
            'title': self.title,
            'dag_id': self.dag_id,
            'dag': {
                'id': self.dag.id,
                'name': self.dag.name,
                'version': self.dag.version
            },
            'state': self.state,
            'state_display': self.get_state_display(),
            'root_id': self.root_id,
            'parent_id': self.parent_id,
            'config': self.config,
            'retries': self.retries,
            'fission_count': self.fission_count,
            'fission_index': self.fission_index,
            'iter_index': self.iter_index,
            'iter_end': self.iter_end,
            'loop_index': self.loop_index,
            'input': self.input,
            'context': self.context,
            'output': self.output,
            'error': self.error,
            'start_time': self.start_time,
            'start_timestamp': self.start_time.timestamp() * 1000 if self.start_time else None,
            'end_time': self.end_time,
            'end_timestamp': self.end_time.timestamp() * 1000 if self.end_time else None,
            'duration': self.duration,
            'create_time': self.create_time,
            'create_timestamp': self.create_time.timestamp() * 1000 if self.create_time else None,
            'update_time': self.update_time,
            'update_timestamp': self.update_time.timestamp() * 1000 if self.update_time else None,
        }

        if with_logs:
            d['logs'] = self.logs

        return d

    def to_json(self, *args, **kwargs):
        d = self.to_dict(*args, **kwargs)
        d.pop('start_time')
        d.pop('end_time')
        d.pop('create_time')
        d.pop('update_time')
        return d

    class Meta:
        managed = True
        db_table = 'seaflow_task'
        verbose_name = '任务'
        verbose_name_plural = verbose_name
        unique_together = ['parent', 'dag', 'fission_index', 'iter_index']


class Step(BaseModel):
    """
    任务步骤
    """

    id = models.AutoField(primary_key=True)
    identifier = models.CharField(max_length=64, db_index=True, null=True)

    name = models.CharField(max_length=64)
    title = models.CharField(max_length=64)
    node = models.ForeignKey('Node', db_constraint=False, on_delete=models.PROTECT)
    root = models.ForeignKey('Task', db_constraint=False, related_name='all_steps', on_delete=models.CASCADE)
    task = models.ForeignKey('Task', db_constraint=False, related_name='steps', on_delete=models.CASCADE)
    state = models.CharField(max_length=64,
                             choices=[(item.name, item.value) for item in StepStates])

    previous_steps = models.ManyToManyField('Step', db_constraint=False, related_name='next_steps',
                                            db_table='seaflow_step_to_step')
    previous_tasks = models.ManyToManyField('Task', db_constraint=False, related_name='next_steps',
                                            db_table='seaflow_task_to_step')
    config = models.JSONField('配置', default=dict)
    retries = models.IntegerField('重试次数', default=0)
    # 分裂
    fission_count = models.IntegerField('分裂数', default=1)
    fission_index = models.IntegerField('分裂序号', db_index=True, default=0)

    # 迭代
    iter_index = models.IntegerField('迭代序号', db_index=True, default=0)
    iter_end = models.BooleanField('迭代末端', null=True)

    # 循环
    loop_index = models.IntegerField('循环序号', default=0)

    input = models.JSONField('输入')
    output = models.JSONField('输出', default=dict)

    extra = models.JSONField(default=dict)

    error = models.TextField('错误信息', default='')
    logs = models.JSONField('执行日志', default=list)
    logfile = models.CharField('s3日志文件路径', max_length=256, default='')
    log_refetch_time = models.DateTimeField(null=True, db_index=True)

    start_time = models.DateTimeField(null=True)
    end_time = models.DateTimeField(null=True)
    duration = models.FloatField(null=True)

    create_time = models.DateTimeField(auto_now_add=True)
    update_time = models.DateTimeField(auto_now=True)

    def __str__(self):
        return '%s%s%s' % (
            self.name,
            ' fission-%s' % self.fission_index if self.node.fissionable else '',
            ' iter-%s' % self.iter_index if self.node.iterable else '',
        )

    def to_dict(self, with_logs=False):
        d = {
            'id': self.id,
            'name': self.name,
            'title': self.title,
            'node_id': self.node_id,
            'node': {
              'id': self.node.id,
              'name': self.node.name
            },
            'root_id': self.root_id,
            'task_id': self.task_id,
            'state': self.state,
            'state_display': self.get_state_display(),
            'config': self.config,
            'retries': self.retries,
            'fission_count': self.fission_count,
            'fission_index': self.fission_index,
            'iter_index': self.iter_index,
            'iter_end': self.iter_end,
            'loop_index': self.loop_index,
            'input': self.input,
            'output': self.output,
            'error': self.error,
            'start_time': self.start_time,
            'start_timestamp': self.start_time.timestamp() * 1000 if self.start_time else None,
            'end_time': self.end_time,
            'end_timestamp': self.end_time.timestamp() * 1000 if self.end_time else None,
            'duration': self.duration,
            'create_time': self.create_time,
            'create_timestamp': self.create_time.timestamp() * 1000 if self.create_time else None,
            'update_time': self.update_time,
            'update_timestamp': self.update_time.timestamp() * 1000 if self.update_time else None,
        }

        if with_logs:
            d['logs'] = self.logs

        return d

    def to_json(self, *args, **kwargs):
        d = self.to_dict(*args, **kwargs)
        d.pop('start_time')
        d.pop('end_time')
        d.pop('create_time')
        d.pop('update_time')
        return d

    class Meta:
        managed = True
        db_table = 'seaflow_step'
        verbose_name = '任务步骤'
        verbose_name_plural = verbose_name
        unique_together = ['task', 'node', 'fission_index', 'iter_index']


class Log(BaseModel):
    """
    日志表
    """

    id = models.AutoField(primary_key=True)
    ref_type = models.CharField(max_length=16, choices=[('TASK', 'Task'), ('STEP', 'Step')]
                                )
    ref_id = models.IntegerField(db_index=True)
    ts = models.FloatField()
    content = models.TextField()

    class Meta:
        managed = True
        db_table = 'seaflow_log'
        verbose_name = '日志'
        verbose_name_plural = verbose_name



"""
Django settings for coe project.

Generated by 'django-admin startproject' using Django 2.2.6.

For more information on this file, see
https://docs.djangoproject.com/en/2.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/2.2/ref/settings/
"""

SECRET_KEY = '2@ttqh(gormw%0gf6fb!%h381p=%+-!^*23qx#m@6nyg4%fh8v'

INSTALLED_APPS = [
    'django_extensions',
    'main',
    'seaflow'
]

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.mysql',
        'NAME': 'test',
        'USER': 'root',
        'PASSWORD': 'chenjian',
        'HOST': '127.0.0.1',
        'PORT': '3306',
        'OPTIONS': {'charset': 'utf8mb4'}
    },
}

# Celery
CELERY_ACCEPT_CONTENT = ['json']
CELERY_RESULT_SERIALIZER = 'json'
CELERY_TASK_SERIALIZER = 'json'
CELERY_BROKER_URL = 'redis://127.0.0.1:6379/6'
CELERY_RESULT_BACKEND = 'redis://127.0.0.1:6379/6'



from . import celery_app
from .utils import get_func


@celery_app.task
def publish_external_step(step_id):
    from .base import SeaflowStep
    SeaflowStep.get(step_id)._publish()


@celery_app.task(bind=True)
def start_carrier_step(self, step_id):
    from .base import SeaflowStep
    SeaflowStep.get(step_id)._carry(self)


@celery_app.task
def apply_root_task(task_id):
    from .base import SeaflowTask
    SeaflowTask.get(task_id).apply(sync=True)


@celery_app.task()
def do_callback(func, event, data):
    """
    :param event: 'TASK/STEP_STATE_<TaskStates/StepStates>
    :param data: json data
    :param func: callback func
    :return:
    """

    get_func(func)(event, data)



import datetime
import hashlib
import importlib
import json
import random
import threading
import time
from enum import Enum

import jsonpath_rw as jsonpath


def singleton_class(post_init=None):
    def _dec(cls):
        def _init(func):
            def __init(self, *args, **kwargs):
                result = func(self, *args, **kwargs)
                if post_init:
                    post_init(self)
                return result

            return __init

        cls.__init__ = _init(cls.__init__)

        def __new__(_cls, *args, **kwargs):
            if not hasattr(cls, "_instance"):
                with cls._instance_lock:
                    if not hasattr(cls, "_instance"):
                        cls._instance = object.__new__(_cls)
            return cls._instance

        cls._instance_lock = threading.Lock()
        cls.__new__ = __new__
        return cls

    return _dec


class Dict(dict):

    def __setattr__(self, key, value):
        self[key] = value

    def __getattr__(self, name):
        try:
            return self.get(name)
        except:
            raise


class ComplexJSONEncoder(json.JSONEncoder):
    """
    encoder for json dumps
    """

    def default(self, obj):
        if isinstance(obj, datetime.datetime):
            return obj.strftime('%Y-%m-%d %H:%M:%S')
        elif isinstance(obj, datetime.date):
            return obj.strftime('%Y-%m-%d')
        elif isinstance(obj, NameComparableEnum):
            return str(obj)
        else:
            return json.JSONEncoder.default(self, obj)


class NameComparableEnum(Enum):

    def __eq__(self, other):
        return self.name == other

    def __str__(self):
        return self.name


class SeaflowException(Exception):

    def __init__(self, *args, **kwargs):
        Exception.__init__(self, *args, **kwargs)


class NotPrintException:
    """
    """


class TimeoutException(SeaflowException, NotPrintException):
    """
    """


class RevokeException(SeaflowException, NotPrintException):
    """
    """


class ExternalActionException(SeaflowException, NotPrintException):
    """
    外部action异常
    """


class ExternalActionFailed(ExternalActionException):
    """
    外部action执行失败
    """

class StateException(SeaflowException):
    """
    状态错误
    """


class ParamException(SeaflowException):
    """
    输入/输出参数错误
    """


class ParamDefinitionException(ParamException):
    """
    输入/输出参数不符合定义
    """

    def __init__(self, *fields):
        super().__init__('invalid fields: %s' % ', '.join(fields))


class ParamAdaptException(ParamException):
    """
    输入/输出参数适配错误
    """

    def __init__(self, *fields):
        super().__init__('failed to adapt fields: %s' % ', '.join(fields))


class SeaflowContext(object):
    """
    """

    def __init__(self, task_id=None, task=None):
        from .models import Task
        self.Task = Task
        self.task = task or self.Task.objects.get(pk=task_id)
        self.context = self.task.context

    def set(self, c):
        """
        局部更新
        :param c: dict
        :return:
        """

        from django_mysql.models.functions import JSONSet
        self.task.update(context=JSONSet('context', {'$.%s' % k: v for k, v in c.items()}), _refresh=False)
        self.reload()

    def get(self):
        """
        :return:
        """
        self.reload()
        return self.context

    def __setitem__(self, key, value):
        self.set({key: value})

    def __getitem__(self, item):
        self.reload()
        return self.context[item]

    def reload(self):
        self.task.refresh_from_db()
        self.context = self.task.context


def tracker_flush(instance, logs):
    instance.update(logs=('%s\n%s' % (instance.logs, logs)).strip())


def get_func(func_name):
    """
    根据func_name反射得到func
    :param func_name:例如apps.api.tasks.request_info
    :return:
    """
    rs = func_name.rsplit('.', 1)
    if len(rs) == 2:
        return getattr(importlib.import_module(rs[0]), rs[1])
    else:
        return eval(func_name)


def generate_identifier(value=None, rand=True):
    _id = hashlib.md5()
    if value:
        _id.update(value.encode())
    if rand:
        _id.update(time.asctime().encode())
        _id.update(str(random.randint(0, 999999999)).encode())
    return _id.hexdigest()


# def jsonpath_to_jsonlogic(o):
#     """
#     {"<": ["$.a", "$.b"]} to {"<": [["var": "a"], ["var", "b"]]}
#     :param o:
#     :return:
#     """
#     o = deepcopy(o)
#
#     def _recursion(_o, parent=None, key=None):
#         """
#         :param _o:
#         :return:
#         """
#         if isinstance(_o, dict):
#             for k, v in _o.items():
#                 _recursion(v, _o, k)
#         elif isinstance(_o, list):
#             for k, v in enumerate(_o):
#                 _recursion(v, _o, k)
#         elif isinstance(_o, str):
#             if _o.startswith('$.'):
#                 parent[key] = ['var', _o[2:]]
#
#     _recursion(o)
#     return o


def del_attrs(o, *attrs):
    for attr in attrs:
        if hasattr(o, attr):
            delattr(o, attr)


# class TimeoutManager(object):
#
#     box = {}
#
#     def register(self, key, expire, callback):
#         self.box[key] = (key, expire, callback)
#
#     def listen(self):
#         pass

class Config(dict):

    def __init__(self, **kwargs):
        for k in kwargs.keys():
            if k in self._keys:
                self[k] = kwargs[k]

    def trim(self):
        for k in list(self.keys()):
            if self[k] is None:
                self.pop(k)
        return self

    @classmethod
    def from_json(cls, o):
        return cls(**o)


class CallbackConfig(Config):
    _keys = ('is_async', 'func')


class TaskConfig(Config):
    _keys = ('countdown', 'max_retries', 'retry_countdown', 'timeout', 'callback')


class StepConfig(Config):
    _keys = ('countdown', 'max_retries', 'retry_countdown', 'timeout', 'callback')


def fission_inputs(inputs, fission_key):
    rs = jsonpath.parse(fission_key).find(inputs)
    assert len(rs) == 1, rs
    return [dict(inputs, **{rs[0].path.fields[0]: x}) for x in rs[0].value]


def iter_inputs(inputs, iter_key):
    """
    :param inputs:
    :param iter_key:
    :return: current_inputs, iter_sequence
    """
    rs = jsonpath.parse(iter_key).find(inputs)
    assert len(rs) == 1, rs
    key = rs[0].path.fields[0]
    sequence = rs[0].value
    return dict(inputs, **{key: sequence[0]}), key, sequence


def merge_outputs_of_tasks(tasks):
    fission_dict = {}
    outputs_list = []
    for t in tasks:
        if t.dag.fissionable:
            if t.dag.id not in fission_dict:
                fission_dict[t.dag.id] = [t.output]
            else:
                fission_dict[t.dag.id].append(t.output)
        else:
            outputs_list.append(t.output)

    for _, v in fission_dict.items():
        outputs_list.append(merge_fission_outputs(*v))

    return merge_outputs(*outputs_list)


def merge_outputs_of_steps(steps):
    fission_dict = {}
    outputs_list = []
    for s in steps:
        if s.node.fissionable:
            if s.node.id not in fission_dict:
                fission_dict[s.node.id] = [s.output]
            else:
                fission_dict[s.node.id].append(s.output)
        else:
            outputs_list.append(s.output)

    for _, v in fission_dict.items():
        outputs_list.append(merge_fission_outputs(*v))

    return merge_outputs(*outputs_list)


def merge_outputs(*outputs):
    """
    相同的key，merge后的值为array
    :param outputs:
    :return:
    """
    result = {}
    for o in outputs:
        for k, v in o.items():
            if k not in result:
                result[k] = [v]
            else:
                result[k].append(v)
    for k, v in result.items():
        if len(v) == 1:
            result[k] = v[0]

    return result


def merge_fission_outputs(*outputs):
    result = {}
    for o in outputs:
        for k, v in o.items():
            if k not in result:
                result[k] = [v]
            else:
                result[k].append(v)

    return result



import setuptools

setuptools.setup(
    name="seaflow",
    version="0.0.1-beta-3",
    author="rainware",
    author_email="kevin90116@gmail.com",
    description="the most free process choreographer",
    url="https://github.com/rainware/seaflow",
    project_urls={
        "Source": "https://github.com/rainware/seaflow",
    },
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        'django>=3.1',
        'mysqlclient>=1.4.4',
        'django-extensions>=3.0.3',
        'celery>=5.2.1',
        'django-celery-results>=2.2.0',
        'redis>=3.5.3',
        'flower>=1.0.0',
        'django-mysql>=3.8.1',
        'timeout-decorator>=0.4.1',
        'jsonpath-rw>=1.4.0',
        'json-logic-qubit>=0.9.1'
    ],
    platforms='any'
)



import json
import os
from celery import Celery

# django settings
os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'settings')

# app definition
app = Celery('main')
app.config_from_object('django.conf:settings', namespace='CELERY')
app.autodiscover_tasks(related_name='actions')

import seaflow
seaflow.set_celery_app(app)


from seaflow.base import Seaflow
from seaflow.consts import StepStates
@Seaflow.action()
def test(self, **params):
    print(json.dumps({'root': self.root.name if self.root else None,
                      'task': self.task.name,
                      'step': self.step.name}))
    if self.step.node.name == 'd1' and self.step.retries < self.step.config.get('max_retries', 0):
        1/0
    return {
        'state': StepStates.SUCCESS,
        'data': {
            'a': [1, 2],
            'o': {'o': 'o'},
            'n': 1,
            's': 's',
            'b': True
        }
    }


